"""Chromosight's stats submodule contains helper function to compute statistical estimators from distributions"""

import numpy as np
import scipy.stats as ss


def fdr_correction(pvals):
    """
    Applies false discovery rate correction via the
    Benjamini-Hochberg procedure to adjust input p-values
    for multiple testing. .

    Parameters
    ----------
    pvals : numpy.array of floats
        Array of uncorrected p-values.

    Returns
    -------
    fdr : numpy.array of floats
        Array of corrected p-values (q-values).
    """
    if pvals is None:
        qvals = None
    else:
        pvals = np.array(pvals)
        # get descending indices
        by_descend = pvals.argsort()[::-1]
        # Get mapping from sorted to original indices
        by_orig = by_descend.argsort()
        # Compute the ratio of n_pval by ranks, which can then be
        # multiplied by p-values to get the critical value
        steps = float(len(pvals)) / np.arange(len(pvals), 0, -1)
        # Compute critical values and find last p-value below above it
        # Set all other to the same value and back to 1 if higher
        qvals = np.minimum(1, np.minimum.accumulate(steps * pvals[by_descend]))
        # Unsort q-values to get original order
        qvals = qvals[by_orig]

    return qvals


def corr_to_pval(corr, n, rho0=0):
    """
    Given a list of Pearson correlation coefficient,
    convert them to two-sided log10 p-values. The p-values
    are computed via the fisher transformation described
    on: https://w.wiki/Ksu

    Parameters
    ----------
    corr : numpy.array
        The Pearson correlation coefficients.
    n : int or numpy.array
        The number of observations used to compute correlation
        coefficients. Can be given as an array of the same size as corr
        to give the number of sample in each coefficient.
    rho0 : float
        The correlation value under h0. We test if corr is
        significantly different from rho0.

    Returns
    -------
    numpy.array
        The array of log10-transformed two-sided p-values,
        same size as corr.
    """
    if isinstance(n, int):
        n = np.repeat(n, corr.shape)
    elif isinstance(n, np.ndarray):
        if n.shape != corr.shape:
            raise ValueError("corr and n must have identical shapes.")
    # Apply Fisher z-transformation on coefficients
    z_score = np.arctanh(corr) - np.arctanh(rho0)
    # Get values of the cumulative standard distribution for each zscore
    # Zscores are all set to negative so as to obtain the tail, resulting
    # p-values are multiplied by two to obtain the two-sided values
    # (distribution is symmetric)
    pvals = 2 * ss.norm().cdf(-np.abs(z_score * np.sqrt(n - 3)))

    return np.log10(pvals)



#!/usr/bin/env python3
# coding: utf-8
"""Chromosight's io submodule contains input/output related functions to load
contact matrices in cool format, and save output patterns coordinates and
windows."""

import pandas as pd
import numpy as np
import pathlib
import sys
import csv
import shutil
from urllib.request import urlopen
import json
from jsonschema import validate, ValidationError
from os.path import join, dirname, isdir
from scipy.sparse import coo_matrix, triu


def load_cool(cool_path):
    """
    Reads a cool file into memory and parses it into a COO sparse matrix
    and an array with the starting bin of each chromosome.
    
    Parameters
    ----------
    cool : str
        Path to the input .cool file.

    Returns
    -------
    mat: scipy.sparse.coo_matrix
        Output sparse matrix in coordinate format
    chroms : pandas.DataFrame
        Table of chromosome information. Each row contains the name, length,
        first and last bin of a chromosome.
    bins : pandas.DataFrame
        Table of genomic bins information. Each row contains the chromosome,
        genomic start and end coordinates of a matrix bin.
    bin_size : int
        Matrix resolution. Corresponds to the number of base pairs per matrix bin.
    """
    try:
        import cooler
    except ImportError:
        print(
            "The cooler package is required to use cool files. Please install it first."
        )
        raise

    c = cooler.Cooler(cool_path)  # pylint: disable=undefined-variable
    if c.binsize is None:
        raise ValueError("The cool file must have equally sized bins")
    mat = c.pixels()[:]
    # Number of fragments  (bins) per chromosome
    bins = c.bins()[:]
    chroms = c.chroms()[:]
    # Number of bins per chromosome
    n_bins = bins.groupby("chrom", sort=False).count().start
    n_bins = n_bins.astype(np.int64)
    # Starting bin of each chromosome
    chrom_start = np.insert(np.array(n_bins), 0, 0)
    # Make chromstart cumulative
    chrom_start = np.cumsum(chrom_start)
    # Make a sparse (COO) matrix from the pixels table
    n = int(max(np.amax(mat.bin1_id), np.amax(mat.bin2_id))) + 1
    shape = (n, n)
    mat = coo_matrix(
        (mat["count"], (mat.bin1_id, mat.bin2_id)),
        shape=shape,
        dtype=np.float64,
    )
    # Only keep upper triangle
    mat = triu(mat)
    bins = bins[["chrom", "start", "end"]]
    chroms["start_bin"] = chrom_start[:-1]
    chroms["end_bin"] = chrom_start[1:]
    return mat, chroms, bins, c.binsize


def load_kernel_config(kernel, custom=False):
    """
    Load a kernel configuration from input JSON file.

    All parameters associated with the kernel along its kernel matrices are
    loaded into a dictionary.

    A kernel config file is a JSON file with the following structure:

    .. code-block:: JSON

    {
        "name": str,
        "kernels": [
            str,
            ...
        ],
        "max_dist": int,
        "min_dist": int,
        "max_iterations": int,
        "max_perc_zero": float,
        "max_perc_undetected": float,
        "pearson": float
        "resolution": int
    }


    The kernels field should contain a list of path to kernel matrices to be
    loaded. These path should be relative to the config file. When loaded, the
    kernel field will contain the target matrices as 2D numpy arrays.

    The kernel matrices files themselves are raw tsv files containing a dense
    matrix of numeric value as read by the numpy.loadtxt function.

    Other fields are:

    * name : Name of the pattern
    * max_dist : maximum distance in basepairs to scan from the diagonal
    * max_iterations: maximum number of scanning iterations to perform
    * max_perc_zero: Maximum percentage of empty (0) pixels to include a pattern
    * max_perc_zero: Maximum percentage of missing (nan) pixels to include a pattern
    * pearson: Increasing this value reduces false positive patterns.
    * resolution: Basepair resolution for the kernel matrix.

    Parameters
    ----------
    kernel : str
        The name of the built-in pattern configuration to load if custom is
        False. Otherwise, the path to the custom JSON configuration file to
        load.
    custom : bool
        Determines if a custom JSON configuration file must be loaded, or if a
        preset configuration is used.

    Returns
    -------
    kernel_config : dict
        A dictionary containing a key: value pair for each parameter as well as
        list of kernel matrices under key 'kernels'.
    """
    # Define a schema for kernel configuration validation
    schema = {
        "type": "object",
        "properties": {
            "name": {"type": "string"},
            "kernels": {"type": "array", "items": {"type": "string"}},
            "min_dist": {"type": "number", "minimum": 0},
            "max_dist": {"type": "number", "minimum": 0},
            "max_iterations": {"type": "number", "minimum": 0},
            "min_separation": {"type": "number", "minimum": 1},
            "max_perc_undetected": {"type": "number", "minimum": 0},
            "max_perc_zero": {"type": "number", "minimum": 0},
            "pearson": {"type": "number"},
            "resolution": {"type": "number"},
        },
        "required": [
            "name",
            "kernels",
            "min_dist",
            "max_dist",
            "max_iterations",
            "min_separation",
            "pearson",
            "resolution",
        ],
    }
    # Custom kernel: use litteral path as config path
    if custom:
        config_path = kernel
    # Preset kernel: Find preset config file matching pattern name
    else:
        # Find chromosight installation directory and get kernel config path
        chromo_dir = pathlib.Path(__file__).parents[1]
        preset_kernel_dir = pathlib.Path(join(chromo_dir, "kernels"))
        # Preset config filename should be {pattern}.json
        config_path = join(preset_kernel_dir, f"{kernel}.json")

    # Try parsing config file
    try:
        with open(config_path, "r") as config:
            kernel_config = json.load(config)
    except FileNotFoundError as e:
        if custom:
            err_msg = f"Error: Kernel configuration file {config_path} does not exist.\n"
        else:
            err_msg = f"Error: No preset configuration for pattern {kernel}.\n"
        sys.stderr.write(err_msg)
        raise e

    # Check that JSON file is valid
    try:
        validate(kernel_config, schema)
    except ValidationError:
        sys.stderr.write("Invalid kernel configuration\n")
        raise

    # Load kernel matrices using path in kernel config
    kernel_matrices = [None] * len(kernel_config["kernels"])
    for i, kernel_path in enumerate(kernel_config["kernels"]):
        kernel_path = join(pathlib.Path(config_path).parent, kernel_path)
        kernel_matrices[i] = np.loadtxt(kernel_path)
    # Replace matrices path by their content in the config dictionary
    kernel_config["kernels"] = kernel_matrices

    return kernel_config


def write_patterns(coords, output_prefix, dec=10):
    """
    Writes coordinates to a text file.

    Parameters
    ----------
    coords : pandas.DataFrame
        Pandas dataframe containing the coordinates and score of
        one detected pattern per row.
    pattern_name : str
        Name of the pattern. Will be the basename of the output
        file.
    output_dir : str
        Output path where the file will be saved.
    dec : int
        Number of decimals to keep in correlation scores and p-values.
    """
    file_path = output_prefix + ".tsv"
    coords.to_csv(file_path, sep="\t", index=None, float_format=f"%.{dec}f")


def save_windows(windows, output_prefix, fmt="json"):
    """
    Write windows surrounding detected patterns to a npy or json file.  The
    file contains a 3D array where windows are piled on axis 0, matrix rows are
    on axis 1 and columns on axis 2.

    Parameters
    ----------
    windows : numpy.array of floats
        3D numpy array with axes (windows, rows, columns).
    output_prefix : str
        Output path where the file will be saved, an extension will be added
        based on the value of "format".
    format : str
        Format in which to save windows. Can be either npy for
        numpy's binary format, or json for a general purpose text
        format.
    """
    if fmt == "npy":
        file_path = output_prefix + ".npy"
        np.save(file_path, windows)
    elif fmt == "json":
        file_path = output_prefix + ".json"
        json_wins = {idx: win.tolist() for idx, win in enumerate(windows)}
        with open(file_path, "w") as handle:
            json.dump(json_wins, handle, indent=4)
    else:
        raise ValueError("window format must be either npy or json.")


def progress(count, total, status=""):
    """
    Basic progress bar in terminal.

    Parameters
    ----------
    count : float
        Current task id.
    total : float
        Maximum task id.
    status : str
        Info to write on the side of the bar.
    """
    bar_len = 20
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = "=" * filled_len + "-" * (bar_len - filled_len)

    # Clear current line, write status and erase everything to the end of
    # the line using ANSI code
    sys.stderr.write("\r [%s] %s%s %s\033[K" % (bar, percents, "%", status))
    sys.stderr.flush()


def load_bed2d(path):
    """
    Loads only the first 6 columns of a 2D BED file. Will sniff for header
    and generate a default header only if none is present.
    Compatible with output of chromosight detect.

    Parameters
    ----------
    path : str
        The path to the 2D BED file.

    Returns
    -------
    pandas.DataFrame :
        The content of the 2D BED file as a dataframe with 6 columns. Header
        will be: chrom1, start1, end1, chrom2, start2, end2.
    """
    header_finder = csv.Sniffer()
    with open(path) as f:
        header = header_finder.has_header(f.read(65536))
    if header:
        bed2d = pd.read_csv(path, sep="\t", header=0, usecols=range(6))
    else:
        bed2d = pd.read_csv(
            path,
            sep="\t",
            header=None,
            names=["chrom1", "start1", "end1", "chrom2", "start2", "end2"],
            usecols=range(6),
        )
    # Make sure start1 is always the left anchor for intrachromosomal pairs
    bed2d.chrom1 = bed2d.chrom1.astype(str)
    bed2d.chrom2 = bed2d.chrom2.astype(str)
    inverted = (bed2d.start2 < bed2d.start1) & (bed2d.chrom1 == bed2d.chrom2)
    bed2d.start1[inverted], bed2d.start2[inverted] = (
        bed2d.start2[inverted],
        bed2d.start1[inverted],
    )
    bed2d.end1[inverted], bed2d.end2[inverted] = (
        bed2d.end2[inverted],
        bed2d.end1[inverted],
    )

    return bed2d


def download_file(url, file, length=16 * 1024):
    req = urlopen(url)
    with open(file, "wb") as fp:
        shutil.copyfileobj(req, fp, length)


def check_prefix_dir(prefix):
    """Checks for existence of the parent directory of an output prefix"""
    out_dir = dirname(prefix)
    if out_dir and not isdir(out_dir):
        raise OSError(f"Directory {out_dir} does not exist.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Pattern exploration and detection

Explore and detect patterns (loops, borders, centromeres, etc.) in Hi-C contact
maps with pattern matching.

Usage:
    chromosight detect  [--kernel-config=FILE] [--pattern=loops]
                        [--pearson=auto] [--win-size=auto] [--iterations=auto]
                        [--win-fmt={json,npy}] [--norm={auto,raw,force}]
                        [--subsample=no] [--inter] [--tsvd] [--smooth-trend]
                        [--n-mads=5] [--min-dist=0] [--max-dist=auto]
                        [--no-plotting] [--min-separation=auto] [--dump=DIR]
                        [--threads=1] [--perc-zero=auto]
                        [--perc-undetected=auto] <contact_map> <prefix>
    chromosight generate-config [--preset loops] [--click contact_map]
                        [--norm={auto,raw,norm}] [--win-size=auto] [--n-mads=5]
                        [--threads=1] <prefix>
    chromosight quantify [--inter] [--pattern=loops] [--subsample=no]
                         [--win-fmt=json] [--kernel-config=FILE] [--norm={auto,raw,norm}]
                         [--threads=1] [--n-mads=5] [--win-size=auto]
                         [--perc-undetected=auto] [--perc-zero=auto]
                         [--no-plotting] [--tsvd] <bed2d> <contact_map> <prefix>
    chromosight list-kernels [--long] [--mat] [--name=kernel_name]
    chromosight test

    detect:
        performs pattern detection on a Hi-C contact map via template matching
    generate-config:
        Generate pre-filled config files to use for detect and quantify.
        A config consists of a JSON file describing parameters for the
        analysis and path pointing to kernel matrices files. Those matrices
        files are tsv files with numeric values as kernel to use for
        convolution.
    quantify:
        Given a list of pairs of positions and a contact map, computes the
        correlation coefficients between those positions and the kernel of the
        selected pattern.
    list-kernels:
        Prints information about available kernels.
    test:
        Download example data and run loop detection on it.

Arguments for detect:
    contact_map                 The Hi-C contact map to detect patterns on, in
                                bedgraph2d or cool format.
    prefix                      Common path prefix used to generate output files.
                                Extensions will be added for each file.

Arguments for quantify:
    bed2d                       Tab-separated text files with columns chrom1, start1
                                end1, chrom2, start2, end2. Each line correspond to
                                a pair of positions (i.e. a position in the matrix).
    contact_map                 Path to the contact map, in bedgraph2d or
                                cool format.
    prefix                      Common path prefix used to generate output files.
                                Extensions will be added for each file.

Arguments for generate-config:
    prefix                      Path prefix for config files. If prefix is a/b,
                                files a/b.json and a/b.1.txt will be generated.
                                If a given pattern has N kernel matrices, N txt
                                files are created they will be named a/b.[1-N].txt.
    -e, --preset=loops          Generate a preset config for the given pattern.
                                Preset configs available are "loops" and
                                "borders". [default: loops]
    -c, --click contact_map     Show input contact map and uses double clicks from
                                user to build the kernel. Warning: memory-heavy,
                                reserve for small genomes or subsetted matrices.

Arguments for list-kernels:
    --name=kernel_name      Only show information related to a particular
                            kernel.[default: all]
    --long                  Show default parameters in addition to kernel names.
    --mat                   Prints an ascii representation of the kernel matrix.

Basic options:
    -h, --help                  Display this help message.
    --version                   Display the program's current version.
    --verbose                   Displays the logo.
    -n, --norm={auto,raw,force} Normalization / balancing behaviour. auto: weights
                                present in the cool file are used. raw: raw contact
                                values are used. force: recompute weights and
                                overwrite existing values. raw[default: auto]
    -I, --inter                 Enable to consider interchromosomal contacts.
                                Warning: Experimental feature with high memory
                                consumption, only use with small matrices.
    -m, --min-dist=auto         Minimum distance from the diagonal (in base pairs).
                                at which detection should operate. [default: auto]
    -M, --max-dist=auto         Maximum distance from the diagonal (in base pairs)
                                for detection. [default: auto]
    -P, --pattern=loops         Which pattern to detect. This will use preset
                                configurations for the given pattern. Possible
                                values are: loops, borders, hairpins and
                                centromeres. [default: loops]
    -p, --pearson=auto          Pearson correlation threshold when detecting patterns
                                in the contact map. Lower values leads to potentially
                                more detections, but more false positives. [default: auto]
    -s, --subsample=INT         If greater than 1, subsample INT contacts from the
                                matrix. If between 0 and 1, subsample a proportion of
                                contacts instead. Useful when comparing matrices with
                                different coverages. [default: no]
    -t, --threads=1             Number of CPUs to use in parallel. [default: 1]
    -u, --perc-undetected=auto  Maximum percentage of non-detectable pixels (nan) in
                                windows allowed to report patterns. [default: auto]
    -z, --perc-zero=auto        Maximum percentage of empty (0) pixels in windows
                                allowed to report patterns. [default: auto]

Advanced options:
    -d, --dump=DIR              Directory where to save matrix dumps during
                                processing and detection. Each dump is saved as
                                a compressed npz of a sparse matrix and can be
                                loaded using scipy.sparse.load_npz.
    -i, --iterations=auto       How many iterations to perform after the first
                                template-based pass. [default: 1]
    -k, --kernel-config=FILE    Optionally give a path to a custom JSON kernel
                                config path. Use this to override pattern if
                                you do not want to use one of the preset
                                patterns.
    --no-plotting               Disable generation of pileup plots.
    -N, --n-mads=5              Maximum number of median absolute deviations below
                                the median of the bin sums distribution allowed to
                                consider detectable bins. [default: 5]
    -S, --min-separation=auto   Minimum distance required between patterns, in
                                basepairs. If two patterns are closer than this
                                distance in both axes, the one with the lowest
                                score is discarded. [default: auto]
    -T, --smooth-trend          Use isotonic regression when detrending to reduce
                                noise at long ranges. Do not enable this for circular
                                genomes.
    -V, --tsvd                  Enable kernel factorisation via truncated svd.
                                Accelerates detection, at the cost of slight
                                inaccuracies. Singular matrices are truncated to
                                retain 99.9% of the information in the kernel.
    -w, --win-fmt={json,npy}    File format used to store individual windows
                                around each pattern. Window order matches
                                patterns inside the associated text file.
                                Possible formats are json and npy. [default: json]
    -W, --win-size=auto         Window size (width), in pixels, to use for the
                                kernel when computing correlations. The pattern
                                kernel will be resized to match this size. Linear
                                linear interpolation is used to fill between pixels.
                                If not specified, the default kernel size will
                                be used instead. [default: auto]


"""
import numpy as np
import pandas as pd
import os
import io
from contextlib import contextmanager
import sys
import json
import docopt
import tempfile
import multiprocessing as mp
from chromosight.version import __version__
from chromosight.utils.contacts_map import HicGenome
import chromosight.utils.io as cio
import chromosight.utils.detection as cid
from chromosight.utils.plotting import (
    pileup_plot,
    click_finder,
    print_ascii_mat,
)
from chromosight.utils.preprocessing import resize_kernel
from chromosight.utils.stats import fdr_correction
import chromosight.kernels as ck
import scipy.ndimage as ndi
import matplotlib.pyplot as plt
import pathlib

LOGO = np.loadtxt(pathlib.Path(__file__).parents[0] / "logo.txt")
URL_EXAMPLE_DATASET = (
    "https://raw.githubusercontent.com/koszullab/"
    "chromosight/master/data_test/example.cool"
)

TEST_LOG = f"""Fetching test dataset at {URL_EXAMPLE_DATASET}...
Running detection on test dataset...
pearson set to 0.3 based on config file.
max_dist set to 2000000 based on config file.
min_dist set to 20000 based on config file.
min_separation set to 5000 based on config file.
max_perc_undetected set to 50.0 based on config file.
max_perc_zero set to 10.0 based on config file.
Matrix already balanced, reusing weights
Preprocessing sub-matrices...
Detecting patterns...
89 patterns detected
Saving patterns in chromosight_test.tsv
Saving patterns in chromosight_test.json
"""


def _override_kernel_config(param_name, param_value, param_type, config):
    """
    Helper function to determine if a config file value should be overriden
    by the user.
    """

    if param_value == "auto":
        try:
            sys.stderr.write(
                f"{param_name} set to {config[param_name]} based on config file.\n"
            )
        except KeyError:
            raise KeyError(
                f"{param_name} is not defined in the config. Please add it to "
                f"the JSON config file, or provide it as a command line option."
            )
    else:
        try:
            config[param_name] = param_type(param_value)
        except ValueError:
            raise ValueError(
                f'Error: {param_name} must be a {param_type} or "auto"'
            )

    return config


def _quantify_sub_mat(data):
    sub = data[0][1]
    config = data[1]
    kernel = data[2]
    positions = data[3]
    sub.contact_map.create_mat()
    # Feed the submatrix to quantification pipeline
    patterns, windows = cid.pattern_detector(
        sub.contact_map,
        config,
        kernel,
        coords=np.array(positions.loc[:, ["bin1", "bin2"]]),
        full=True,
        tsvd=config["tsvd"],
    )
    sub.contact_map.destroy_mat()

    return {
        "coords": patterns,
        "windows": windows,
        "chr1": sub.chr1,
        "chr2": sub.chr2,
    }


def _get_chrom_pos(positions, hic_genome, chr1, chr2):
    # Filter patterns falling onto this sub-matrix
    sub_pat = positions.loc[
        (positions.chrom1 == chr1) & (positions.chrom2 == chr2)
    ]
    # Convert genomic coordinates to bins for horizontal and vertical axes
    for ax in [1, 2]:
        sub_pat_ax = sub_pat.loc[:, [f"chrom{ax}", f"pos{ax}"]].rename(
            columns={f"chrom{ax}": "chrom", f"pos{ax}": "pos"}
        )
        sub_pat_bins = hic_genome.coords_to_bins(sub_pat_ax)
        sub_pat[f"bin{ax}"] = sub_pat_bins

    # Check for nan bins (coords that do not match any Hi-C fragments
    fall_out = np.isnan(sub_pat["bin1"]) | np.isnan(sub_pat["bin2"])
    if np.any(fall_out):
        n_out = len(sub_pat_bins[fall_out])
        sys.stderr.write(
            f"\n{n_out} entr{'ies' if n_out > 1 else 'y'} outside "
            "genomic coordinates of the Hi-C matrix will be ignored.\n"
        )
        sub_pat = sub_pat.loc[~fall_out, :]
    sub_pat_idx = sub_pat.index.values
    # Convert bins from whole genome matrix to sub matrix
    sub_pat = hic_genome.get_sub_mat_pattern(chr1, chr2, sub_pat)
    return sub_pat_idx, sub_pat


def cmd_quantify(args):
    bed2d_path = args["<bed2d>"]
    mat_path = args["<contact_map>"]
    prefix = args["<prefix>"]
    n_mads = float(args["--n-mads"])
    pattern = args["--pattern"]
    inter = args["--inter"]
    kernel_config_path = args["--kernel-config"]
    perc_zero = args["--perc-zero"]
    perc_undetected = args["--perc-undetected"]
    plotting_enabled = False if args["--no-plotting"] else True
    threads = int(args["--threads"])
    norm = args["--norm"]
    tsvd = 0.999 if args["--tsvd"] else None
    win_fmt = args["--win-fmt"]
    if win_fmt not in ["npy", "json"]:
        sys.stderr.write("Error: --win-fmt must be either json or npy.\n")
        sys.exit(1)
    win_size = args["--win-size"]
    if win_size != "auto":
        win_size = int(win_size)
    subsample = args["--subsample"]
    # If prefix involves a directory, crash if it does not exist
    cio.check_prefix_dir(prefix)
    # Load 6 cols from 2D BED file and infer header
    bed2d = cio.load_bed2d(bed2d_path)
    # Warn user if --inter is disabled but list contains inter patterns
    if not inter and len(bed2d.start1[bed2d.chrom1 != bed2d.chrom2]) > 0:
        sys.stderr.write(
            "Warning: The bed2d file contains interchromosomal patterns. "
            "These patterns will not be scanned unless --inter is used.\n"
        )
    if kernel_config_path is not None:
        custom = True
        # Loading input path as config
        config_path = kernel_config_path
    else:
        custom = False
        # Will use a preset config file matching pattern name
        config_path = pattern
    cfg = cio.load_kernel_config(config_path, custom)
    # Subsample Hi-C contacts from the matrix, if requested
    if subsample == "no":
        subsample = None
    # Instantiate and preprocess contact map
    hic_genome = HicGenome(
        mat_path, inter=inter, kernel_config=cfg, sample=subsample
    )
    # enforce max scanning distance to pattern at longest distance
    furthest = np.max(bed2d.start2 - bed2d.start1)
    max_diag = hic_genome.clr.shape[0] * hic_genome.clr.binsize
    cfg["max_dist"] = min(furthest, max_diag)
    cfg["min_dist"] = 0
    cfg["tsvd"] = tsvd
    cfg = _override_kernel_config("max_perc_zero", perc_zero, float, cfg)
    cfg = _override_kernel_config(
        "max_perc_undetected", perc_undetected, float, cfg
    )

    # Notify contact map instance of changes in scanning distance
    hic_genome.kernel_config = cfg
    # Normalize (balance) matrix using ICE
    hic_genome.normalize(norm=norm, n_mads=n_mads, threads=threads)
    # Initialize output structures
    bed2d["score"] = np.nan
    bed2d["pvalue"] = np.nan
    positions = bed2d.copy()
    # Only resize kernel matrix if explicitely requested
    km, kn = cfg["kernels"][0].shape
    if win_size != "auto":
        if not win_size % 2:
            raise ValueError("--win-size must be odd")
        for i, k in enumerate(cfg["kernels"]):
            cfg["kernels"][i] = resize_kernel(k, factor=win_size / km)
        km = kn = win_size
        # Update kernel config after resizing kernels
        hic_genome.kernel_config = cfg
    # Define how many diagonals should be used in intra-matrices
    hic_genome.compute_max_dist()
    # Split whole genome matrix into intra- and inter- sub matrices. Each sub
    # matrix is processed on the fly (obs / exp, trimming diagonals > max dist)
    hic_genome.make_sub_matrices()
    windows = np.full((positions.shape[0], km, kn), np.nan)
    # For each position, we use the center of the BED interval
    positions["pos1"] = (positions.start1 + positions.end1) // 2
    positions["pos2"] = (positions.start2 + positions.end2) // 2
    # Use each kernel matrix available for the pattern
    for kernel_id, kernel_matrix in enumerate(cfg["kernels"]):
        cio.progress(kernel_id, len(cfg["kernels"]), f"Kernel: {kernel_id}\n")
        # Iterate over intra- and inter-chromosomal sub-matrices
        n_sub_mats = hic_genome.sub_mats.shape[0]
        # Apply quantification procedure to all sub matrices in parallel
        sub_pos = [
            _get_chrom_pos(positions, hic_genome, m[1].chr1, m[1].chr2)
            for m in hic_genome.sub_mats.iterrows()
        ]
        sub_mat_data = zip(
            hic_genome.sub_mats.iterrows(),
            [cfg for _ in range(n_sub_mats)],
            [kernel_matrix for _ in range(n_sub_mats)],
            [s[1] for s in sub_pos],
        )
        # Run quantification in parallel on different sub matrices, and show progress when
        # gathering results
        sub_mat_results = []
        # Run in multiprocessing subprocesses
        if threads > 1:
            pool = mp.Pool(threads)
            dispatcher = pool.imap(_quantify_sub_mat, sub_mat_data, 1)
        else:
            dispatcher = map(_quantify_sub_mat, sub_mat_data)
        for s, result in enumerate(dispatcher):
            cio.progress(s, n_sub_mats, f"{result['chr1']}-{result['chr2']}")
            sub_mat_results.append(result)

        for i, r in enumerate(sub_mat_results):
            sub_pat_idx = sub_pos[i][0]

            # For each coordinate, keep the highest coefficient
            # among all kernels.
            try:
                if kernel_id == 0:
                    bed2d["score"][sub_pat_idx] = r["coords"].score.values
                    bed2d["pvalue"][sub_pat_idx] = r["coords"].pvalue.values
                    windows[sub_pat_idx, :, :] = r["windows"]
                else:
                    # Only update scores and their corresponding windows
                    # if better than results from previous kernels
                    better = (
                        r["coords"].score > bed2d["score"][sub_pat_idx].values
                    ) | (np.insnan(bed2d["score"][sub_pat_idx].values))
                    better_idx = sub_pat_idx[better]
                    bed2d["score"][better_idx] = (
                        r["coords"].score[better].values
                    )
                    windows[better_idx, :, :] = r["windows"][better]
            # Do nothing if no pattern was detected or matrix
            # is smaller than the kernel (-> patterns is None)
            except AttributeError:
                pass
        bed2d["bin1"] = hic_genome.coords_to_bins(
            bed2d.loc[:, ["chrom1", "start1"]].rename(
                columns={"chrom1": "chrom", "start1": "pos"}
            )
        )
        bed2d["bin2"] = hic_genome.coords_to_bins(
            bed2d.loc[:, ["chrom2", "start2"]].rename(
                columns={"chrom2": "chrom", "start2": "pos"}
            )
        )
        bed2d["qvalue"] = fdr_correction(bed2d["pvalue"])
        bed2d = bed2d.loc[
            :,
            [
                "chrom1",
                "start1",
                "end1",
                "chrom2",
                "start2",
                "end2",
                "bin1",
                "bin2",
                "score",
                "pvalue",
                "qvalue",
            ],
        ]
        # Set p-values of invalid scores to nan
        bed2d.loc[np.isnan(bed2d.score), "pvalue"] = np.nan
        bed2d.loc[np.isnan(bed2d.score), "qvalue"] = np.nan
        cio.write_patterns(bed2d, prefix)
        cio.save_windows(windows, prefix, fmt=win_fmt)
        # Generate pileup visualisations if requested
        if plotting_enabled:
            # Compute and plot pileup
            pileup_title = ("pileup_of_{n}_{pattern}").format(
                pattern=cfg["name"], n=windows.shape[0]
            )
            windows_pileup = cid.pileup_patterns(windows)
            # Symmetrize pileup for diagonal patterns
            if not cfg["max_dist"]:
                # Replace nan below diag by 0
                windows_pileup = np.nan_to_num(windows_pileup)
                # Add transpose
                windows_pileup += np.transpose(windows_pileup) - np.diag(
                    np.diag(windows_pileup)
                )
            sys.stderr.write(f"Saving pileup plots in {prefix}.pdf\n")
            pileup_plot(windows_pileup, prefix, name=pileup_title)


def cmd_generate_config(args):
    # Parse command line args for generate_config
    prefix = args["<prefix>"]
    pattern = args["--preset"]
    click_find = args["--click"]
    n_mads = float(args["--n-mads"])
    norm = args["--norm"]
    win_size = args["--win-size"]
    threads = int(args["--threads"])

    cfg = cio.load_kernel_config(pattern, False)

    # If prefix involves a directory, crash if it does not exist
    cio.check_prefix_dir(prefix)

    # If a specific window size if requested, resize all kernels
    if win_size != "auto":
        win_size = int(win_size)
        if not win_size % 2:
            raise ValueError("--win-size must be odd")
        resize = lambda m: resize_kernel(m, factor=win_size / m.shape[0])
        cfg["kernels"] = [resize(k) for k in cfg["kernels"]]
    # Otherwise, just inherit window size from the kernel config
    else:
        win_size = cfg["kernels"][0].shape[0]

    # If click mode is enabled, build a kernel from scratch using
    # graphical display, otherwise, just inherit the pattern's kernel
    if click_find:
        hic_genome = HicGenome(click_find, inter=True, kernel_config=cfg)
        # Normalize (balance) the whole genome matrix
        hic_genome.normalize(norm=norm, n_mads=n_mads, threads=threads)
        # enforce full scanning distance in kernel config

        hic_genome.max_dist = hic_genome.clr.shape[0] * hic_genome.clr.binsize
        # Process each sub-matrix individually (detrend diag for intra)
        hic_genome.make_sub_matrices()
        for sub in hic_genome.sub_mats.iterrows():
            sub_mat = sub[1]
            sub_mat.contact_map.create_mat()
        processed_mat = hic_genome.gather_sub_matrices().tocsr()
        windows = click_finder(processed_mat, half_w=int((win_size - 1) / 2))
        # Pileup all recorded windows and convert to JSON serializable list
        pileup = ndi.gaussian_filter(cid.pileup_patterns(windows), 1)
        cfg["kernels"] = [pileup.tolist()]
        # Show the newly generate kernel to the user, use zscore to highlight contrast
        hm = plt.imshow(
            np.log(pileup), vmax=np.percentile(pileup, 99), cmap="afmhot_r"
        )
        cbar = plt.colorbar(hm)
        cbar.set_label("Log10 Hi-C contacts")
        plt.title("Manually generated kernel")
        plt.show()
    # Write kernel matrices to files with input prefix and replace kernels
    # by their path in config
    for mat_id, mat in enumerate(cfg["kernels"]):
        mat_path = f"{prefix}.{mat_id+1}.txt"
        np.savetxt(mat_path, mat)
        cfg["kernels"][mat_id] = mat_path

    # Write config to JSON file using prefix
    with open(f"{prefix}.json", "w") as config_handle:
        json.dump(cfg, config_handle, indent=4)


def _detect_sub_mat(data):
    sub = data[0][1]
    config = data[1]
    kernel = data[2]
    dump = data[3]
    sub.contact_map.create_mat()
    chrom_patterns, chrom_windows = cid.pattern_detector(
        sub.contact_map,
        config,
        kernel,
        dump=dump,
        full=True,
        tsvd=config["tsvd"],
    )
    sub.contact_map.destroy_mat()

    return {
        "coords": chrom_patterns,
        "windows": chrom_windows,
        "chr1": sub.chr1,
        "chr2": sub.chr2,
    }


def cmd_detect(args):
    # Parse command line arguments for detect
    dump = args["--dump"]
    norm = args["--norm"]
    interchrom = args["--inter"]
    iterations = args["--iterations"]
    kernel_config_path = args["--kernel-config"]
    mat_path = args["<contact_map>"]
    max_dist = args["--max-dist"]
    min_dist = args["--min-dist"]
    min_separation = args["--min-separation"]
    n_mads = float(args["--n-mads"])
    prefix = args["<prefix>"]
    pattern = args["--pattern"]
    pearson = args["--pearson"]
    perc_zero = args["--perc-zero"]
    perc_undetected = args["--perc-undetected"]
    subsample = args["--subsample"]
    threads = int(args["--threads"])
    tsvd = 0.999 if args["--tsvd"] else None
    win_fmt = args["--win-fmt"]
    win_size = args["--win-size"]
    if subsample == "no":
        subsample = None
    plotting_enabled = False if args["--no-plotting"] else True
    smooth_trend = args["--smooth-trend"]
    if smooth_trend is None:
        smooth_trend = False

    # If prefix involves a directory, crash if it does not exist
    cio.check_prefix_dir(prefix)

    if win_fmt not in ["npy", "json"]:
        sys.stderr.write("Error: --win-fmt must be either json or npy.\n")
        sys.exit(1)
    # Read a user-provided kernel config if custom is true
    # Else, load a preset kernel config for input pattern
    # Configs are JSON files containing all parameter associated with the pattern
    # They are loaded into a dictionary in the form :
    # {"max_iterations": 3, "kernels": [kernel1, kernel2, ...], ...}
    # Where each kernel is a 2D numpy array representing the pattern
    if kernel_config_path is not None:
        custom = True
        # Loading input path as config
        config_path = kernel_config_path
    else:
        custom = False
        # Will use a preset config file matching pattern name
        config_path = pattern

    ### 0: LOAD INPUT
    params = {
        "max_iterations": (iterations, int),
        "pearson": (pearson, float),
        "max_dist": (max_dist, int),
        "min_dist": (min_dist, int),
        "min_separation": (min_separation, int),
        "max_perc_undetected": (perc_undetected, float),
        "max_perc_zero": (perc_zero, float),
    }
    cfg = cio.load_kernel_config(config_path, custom)
    for param_name, (param_value, param_type) in params.items():
        cfg = _override_kernel_config(param_name, param_value, param_type, cfg)

    # Resize kernels if requested
    if win_size != "auto":
        win_size = int(win_size)
        if not win_size % 2:
            raise ValueError("--win-size must be odd")
        resize = lambda m: resize_kernel(m, factor=win_size / m.shape[0])
        cfg["kernels"] = [resize(k) for k in cfg["kernels"]]

    if interchrom:
        sys.stderr.write(
            "WARNING: Detection on interchromosomal matrices is expensive in RAM\n"
        )
    hic_genome = HicGenome(
        mat_path,
        inter=interchrom,
        kernel_config=cfg,
        dump=dump,
        smooth=smooth_trend,
        sample=subsample,
    )
    ### 1: Process input signal
    hic_genome.kernel_config = cfg
    # Normalize (balance) matrix using ICE
    hic_genome.normalize(norm=norm, n_mads=n_mads, threads=threads)
    # Define how many diagonals should be used in intra-matrices
    hic_genome.compute_max_dist()
    # Split whole genome matrix into intra- and inter- sub matrices. Each sub
    # matrix is processed on the fly (obs / exp, trimming diagonals > max dist)
    hic_genome.make_sub_matrices()

    all_coords = []
    all_windows = []

    ### 2: DETECTION ON EACH SUBMATRIX
    n_sub_mats = hic_genome.sub_mats.shape[0]
    # Loop over the different kernel matrices for input pattern
    run_id = 0
    # Use cfg to inform jobs whether they should run full convolution
    cfg["tsvd"] = tsvd
    total_runs = len(cfg["kernels"]) * cfg["max_iterations"]
    sys.stderr.write("Detecting patterns...\n")
    for kernel_id, kernel_matrix in enumerate(cfg["kernels"]):
        # Adjust kernel iteratively
        for i in range(cfg["max_iterations"]):
            cio.progress(
                run_id, total_runs, f"Kernel: {kernel_id}, Iteration: {i}\n"
            )

            # Apply detection procedure to all sub matrices in parallel
            sub_mat_data = zip(
                hic_genome.sub_mats.iterrows(),
                [cfg for i in range(n_sub_mats)],
                [kernel_matrix for i in range(n_sub_mats)],
                [dump for i in range(n_sub_mats)],
            )
            # Run detection in parallel on different sub matrices, and show progress when
            # gathering results
            sub_mat_results = []
            # Run in multiprocessing subprocesses
            if threads > 1:
                pool = mp.Pool(threads)
                dispatcher = pool.imap(_detect_sub_mat, sub_mat_data, 1)
            else:
                dispatcher = map(_detect_sub_mat, sub_mat_data)
            for s, result in enumerate(dispatcher):
                cio.progress(s, n_sub_mats, f"{result['chr1']}-{result['chr2']}")
                sub_mat_results.append(result)

            # Convert coordinates from chromosome to whole genome bins
            kernel_coords = [
                hic_genome.get_full_mat_pattern(
                    d["chr1"], d["chr2"], d["coords"]
                )
                for d in sub_mat_results
                if d["coords"] is not None
            ]

            # Gather newly detected pattern coordinates
            try:
                # Extract surrounding windows for each sub_matrix
                kernel_windows = np.concatenate(
                    [
                        w["windows"]
                        for w in sub_mat_results
                        if w["windows"] is not None
                    ],
                    axis=0,
                )
                all_coords.append(
                    pd.concat(kernel_coords, axis=0).reset_index(drop=True)
                )
                # Add info about kernel and iteration which detected these patterns
                all_coords[-1]["kernel_id"] = kernel_id
                all_coords[-1]["iteration"] = i
                all_windows.append(kernel_windows)

            # If no pattern was found with this kernel
            # skip directly to the next one, skipping iterations
            except ValueError:
                break

            # Update kernel with patterns detected at current iteration
            kernel_matrix = cid.pileup_patterns(kernel_windows)
            run_id += 1
    cio.progress(run_id, total_runs, f"Kernel: {kernel_id}, Iteration: {i}\n")
    # If no pattern detected on any chromosome, with any kernel, exit gracefully
    if len(all_coords) == 0:
        sys.stderr.write("No pattern detected ! Exiting.\n")
        sys.exit(0)
    # Finish parallelized part
    if threads > 1:
        pool.close()
    # Combine patterns of all kernel matrices into a single array
    all_coords = pd.concat(all_coords, axis=0).reset_index(drop=True)
    # Combine all windows from different kernels into a single pile of windows
    all_windows = np.concatenate(all_windows, axis=0)

    # Compute minimum separation in bins and make sure it has a reasonable value
    separation_bins = int(cfg["min_separation"] // hic_genome.clr.binsize)
    if separation_bins < 1:
        separation_bins = 1
    print(f"Minimum pattern separation is : {separation_bins}")
    # Remove patterns with overlapping windows (smeared patterns)
    distinct_patterns = cid.remove_neighbours(
        all_coords, win_size=separation_bins
    )

    # Drop patterns that are too close to each other
    all_coords = all_coords.loc[distinct_patterns, :]
    all_windows = all_windows[distinct_patterns, :, :]

    # Get from bins into basepair coordinates
    coords_1 = hic_genome.bins_to_coords(all_coords.bin1).reset_index(
        drop=True
    )
    coords_1.columns = [str(col) + "1" for col in coords_1.columns]
    coords_2 = hic_genome.bins_to_coords(all_coords.bin2).reset_index(
        drop=True
    )
    coords_2.columns = [str(col) + "2" for col in coords_2.columns]

    all_coords = pd.concat(
        [all_coords.reset_index(drop=True), coords_1, coords_2], axis=1
    )

    # Filter patterns closer than minimum distance from the diagonal if any
    min_dist_drop_mask = (all_coords.chrom1 == all_coords.chrom2) & (
        np.abs(all_coords.start2 - all_coords.start1) < cfg["min_dist"]
    )
    all_coords = all_coords.loc[~min_dist_drop_mask, :]
    all_windows = all_windows[~min_dist_drop_mask, :, :]

    # Correct p-values for multiple testing using FDR
    all_coords["qvalue"] = fdr_correction(all_coords["pvalue"])
    # Reorder columns
    all_coords = all_coords.loc[
        :,
        [
            "chrom1",
            "start1",
            "end1",
            "chrom2",
            "start2",
            "end2",
            "bin1",
            "bin2",
            "kernel_id",
            "iteration",
            "score",
            "pvalue",
            "qvalue",
        ],
    ]

    ### 3: WRITE OUTPUT
    sys.stderr.write(f"{all_coords.shape[0]} patterns detected\n")
    # Save patterns and their coordinates in a tsv file

    sys.stderr.write(f"Saving patterns in {prefix}.tsv\n")
    cio.write_patterns(all_coords, prefix)
    # Save windows as an array in an npy file

    sys.stderr.write(f"Saving patterns in {prefix}.{win_fmt}\n")
    cio.save_windows(all_windows, prefix, fmt=win_fmt)

    # Generate pileup visualisations if requested
    if plotting_enabled:
        # Compute and plot pileup
        pileup_title = ("Pileup of {n} {pattern}").format(
            pattern=cfg["name"], n=all_windows.shape[0]
        )
        windows_pileup = cid.pileup_patterns(all_windows)
        # Symmetrize pileup for diagonal patterns
        if not cfg["max_dist"]:
            # Replace nan below diag by 0
            windows_pileup = np.nan_to_num(windows_pileup)
            # Add transpose
            windows_pileup += np.transpose(windows_pileup) - np.diag(
                np.diag(windows_pileup)
            )
        sys.stderr.write(f"Saving pileup plots in {prefix}.pdf\n")
        pileup_plot(windows_pileup, prefix, name=pileup_title)


def cmd_list_kernels(args):

    kernel_name = args["--name"]
    # Load every avaiable kernel by default
    if kernel_name == "all":
        kernels = ck.kernel_names
    # If a specific kernel was requested, only load this one
    else:
        kernels = [kernel_name]

    # Check availability of each kernel and print its name
    for k in kernels:
        try:
            kernel_infos = getattr(ck, k)
        except AttributeError:
            raise ValueError(f"Kernel {k} is not available")
        print(k)
        # Print default params if --long specified (key-value pairs in json)
        if args["--long"]:
            exclude_params = ["name", "resolution", "kernels"]
            for param, value in kernel_infos.items():
                if param not in exclude_params:
                    print(f"  {param}: {value}")
        if args["--mat"]:
            mats = kernel_infos["kernels"]
            for mat in mats:
                print_ascii_mat(mat)


def cmd_test(args):

    sys.stderr.write(f"Fetching test dataset at {URL_EXAMPLE_DATASET}...\n")
    tmp_cool = tempfile.NamedTemporaryFile(delete=False)
    cio.download_file(URL_EXAMPLE_DATASET, tmp_cool.name)

    sys.stderr.write(f"Running detection on test dataset...\n")

    args["<contact_map>"] = tmp_cool.name
    args["<prefix>"] = "chromosight_test"
    args["--no-plotting"] = True
    cmd_detect(args)
    os.unlink(tmp_cool.name)


@contextmanager
def capture_ouput(stderr_to=None):
    """Capture the stderr of the test run. """

    try:
        stderr = sys.stderr
        sys.stderr = c2 = stderr_to or io.StringIO()
        yield c2

    finally:
        sys.stderr = stderr
        try:
            c2.flush()
            c2.seek(0)
        except (ValueError, IOError):
            pass


def logo_version(logo, ver):
    small_logo = resize_kernel(logo, factor=0.33, quiet=True)
    ascii_logo = print_ascii_mat(small_logo, colored=False, print_str=False)
    return f"{ascii_logo} Chromosight version {ver}"


def main():

    args = docopt.docopt(__doc__, version=logo_version(LOGO, __version__))
    detect = args["detect"]
    generate_config = args["generate-config"]
    list_kernels = args["list-kernels"]
    quantify = args["quantify"]
    test = args["test"]
    if test:
        with capture_ouput() as stderr:
            cmd_test(args)

        obs_log = stderr.read()
        sys.stderr.write(obs_log)

        # remove progress bars and \r chars
        obs_log_lines = {
            u.strip("\x1b[K")
            for u in set(obs_log.split("\n"))
            if "\r" not in u
        }
        exp_log_lines = set(TEST_LOG.split("\n"))

        if len(exp_log_lines ^ obs_log_lines):
            sys.stderr.write(
                "\nWarning, the test log differed from the "
                "expected one. This means the program changed its output from"
                "previous versions. You may ignore this if you are not a "
                "developer.\n\n"
                f"Here is the expected log:\n\n{TEST_LOG}\n"
            )

    elif detect:
        cmd_detect(args)
    elif generate_config:
        cmd_generate_config(args)
    elif list_kernels:
        cmd_list_kernels(args)
    elif quantify:
        cmd_quantify(args)

    return 0


if __name__ == "__main__":
    main()



"""Chromosight's plotting submodule contains utilities to visualize the pileup
of detected patterns or the input matrix. It also implements an interactive map
recording the coordinates of double clicks."""

import os
import sys
import numpy as np
from matplotlib import pyplot as plt


def pileup_plot(pileup_pattern, output_prefix, name="pileup_patterns"):
    """
    Wrapper around matplotlib.pyplot.imshow to visualize the pileup of patterns
    detected by chromosight
    """

    plt.imshow(
        pileup_pattern,
        interpolation="none",
        vmin=0.0,
        vmax=2.0,
        cmap="seismic",
    )
    plt.title("{} pileup".format(name))
    plt.colorbar()
    plt.savefig(output_prefix + ".pdf", dpi=100, format="pdf")
    plt.close("all")


def plot_whole_matrix(
    clr, patterns, out=None, region=None, region2=None, log_transform=False
):
    """
    Visualise the input matrix with a set of patterns overlaid on top.
    Can optionally restrict the visualisation to a region.

    Parameters
    ----------
    mat : scipy.sparse.csr_matrix
        The whole genome Hi-C matrix to be visualized.
    patterns : pandas.DataFrame
        The set of patterns to be plotted on top of the matrix. One pattern per
        row, 3 columns: bin1, bin2 and score of types int, int and float,
        respectively.
    region : str
        The genomic range, in UCSC format, corresponding to rows to be plotted
        in the matrix. If not given, the whole matrix is used. It only region
        is given, but not region2, the matrix is subsetted on rows and columns
        to show a region on the diagonal.
    region2 : str
        The genomic range, in UCSC format, of columns to be plotted in the matrix.
        Region must also be provided, or this will be ignored.
    log_transform : bool
        Whether to log transform the matrix.

    """

    if region is not None and region2 is None:
        mat = clr.matrix().fetch(region)
        s1, e1 = clr.extent(region)
        s2, e2 = s1, e1
    elif region is not None and region2 is not None:
        mat = clr.matrix().fetch(region, region2)
        s1, e1 = clr.extent(region)
        s2, e2 = clr.extent(region2)
    else:
        mat = clr.matrix()[:]
        s1, e1 = 0, clr.shape[0]
        s2, e2 = 0, clr.shape[1]

    pat = patterns.copy()
    pat = pat.loc[
        (pat.bin1 > s1) & (pat.bin1 < e1) & (pat.bin2 > s2) & (pat.bin2 < e2),
        :,
    ]

    if log_transform:
        mat = np.log(mat)
    mat[mat == 0] = np.nan
    plt.figure(dpi=1200)
    plt.imshow(
        mat,
        cmap="Reds",
        vmax=np.percentile(mat[~np.isnan(mat)], 99),
    )
    plt.scatter(
        pat.bin1 - s1,
        pat.bin2 - s2,
        facecolors="none",
        edgecolors="blue",
        s=0.05,
    )
    if out is None:
        plt.show()
    else:
        plt.savefig(out)


def click_finder(mat, half_w=8):
    """
    Given an input Hi-C matrix, show an interactive window and record
    coordinates where the user double-clicks. When the interactive window is
    closed, the stack of windows around recorded coordinates is returned.

    Parameters
    ----------
    mat : scipy.sparse.csr_matrix
        The input Hi-C matrix to display interactively.
    half_w : int
        Half width of the windows to return. The resulting windows

    Returns
    -------
    numpy.array :
        3D stack of images around coordinates recorded interactively. The shape
        of the stack is (N, w, w) where N is the number of coordinates and w is
        2*half_w.
    """
    global coords
    coords = []

    def onclick(event):
        global ix, iy
        global coords
        try:
            ix, iy = int(event.xdata), int(event.ydata)
        except TypeError:
            return None
        try:
            if coords[-1] == (ix, iy):
                print(f"x = {ix}, y = {iy}")
        except IndexError:
            pass
        coords.append((ix, iy))
        return coords

    fig = plt.figure()
    plt.imshow(mat.toarray(), cmap="afmhot_r", vmax=np.percentile(mat.data, 95))
    plt.title("Double click to record pattern positions")
    # Setup click listener
    cid = fig.canvas.mpl_connect("button_press_event", onclick)
    plt.show()
    fig.canvas.mpl_disconnect(cid)
    # Keep coordinates that were clicked twice consecutively (double-click)
    double_clicked = set()
    for c in range(1, len(coords)):
        if coords[c - 1] == coords[c]:
            double_clicked.add(coords[c])
    # initialize empty image stack, will store windows around coords
    img_stack = np.zeros((len(double_clicked), half_w * 2 + 1, half_w * 2 + 1))
    bad_coords = np.zeros(len(double_clicked), dtype=bool)
    # Fill the image stack with windows around each coord
    for i, (center_v, center_h) in enumerate(double_clicked):
        high, low = center_h - half_w, center_h + half_w + 1
        left, right = center_v - half_w, center_v + half_w + 1
        try:
            img_stack[i] = mat[high:low, left:right].toarray()
        except ValueError:
            bad_coords[i] = True
            sys.stderr.write(
                f"Discarding {(center_v, center_h)}: Too close "
                "to the edge of the matrix\n"
            )
    # Discard images associated with coords too close to the edge
    img_stack = img_stack[~bad_coords]
    return img_stack


def print_ascii_mat(mat, adjust=True, colored=False, print_str=True):
    """
    Given a 2D numpy array of float, print it in ASCII art.

    Parameters
    ----------
    mat : np.array of floats
        Matrix to visualize.
    adjust : bool
        Whether to adjust the drawing size to termina width.
    colored : bool
        Whether to use colors.
    print_str : bool
        If true, the ASCII art is printed to stdout, otherwise it
        is stored in a string and returned.

    Returns
    -------
    str :
        An empty string is returned if print_str is True, otherwise the
        ASCII art is returned as a string.
    """

    if adjust:
        try:
            term_width = (os.get_terminal_size()[0] // 2) - 5
        except OSError:
            term_width = 79  # default terminal width fallback
        step = int(max(1, np.ceil(mat.shape[1] / term_width)))
    else:
        step = 1
    ascii_str = " .,:;ox%#@"
    ascii_colors = [
        "\x1b[37m",
        "\x1b[37m",
        "\x1b[36m",
        "\x1b[36m",
        "\x1b[32m",
        "\x1b[32m",
        "\x1b[34m",
        "\x1b[34m",
        "\x1b[33m",
        "\x1b[31m",
    ]
    if colored:
        suffix = "\x1b[0m"
    else:
        suffix = ""
    ascii_art = ""

    def ascii_printer(text, end="\n", print_str=print_str):
        if print_str:
            print(text, end=end)
            out = ''
        else:
            out = text + end
        return out

    sorted_pixels = np.sort(mat.flatten())
    perc_pixels = np.searchsorted(sorted_pixels, mat) / len(sorted_pixels)
    perc_pixels = (10 * perc_pixels).astype(int)
    ascii_art += ascii_printer("  " + "- " * (1 + perc_pixels.shape[1] // step))
    for i in range(0, mat.shape[0], step):
        ascii_art += ascii_printer("  |", end="")
        for j in range(0, mat.shape[1], step):  # pixels are skipped
            pix = perc_pixels[i, j]
            prefix = ascii_colors[pix] if colored else ""
            ascii_art += ascii_printer(f"{prefix}{ascii_str[pix]}{suffix} ", end="")
        ascii_art += ascii_printer("|")
    ascii_art += ascii_printer("  " + "- " * (1 + perc_pixels.shape[1] // step))

    return ascii_art



"""Chromosight's detection submodule implements the bulk of chromosight
convolution engine, as well as functions to perform the different steps of the
detection algorithm (pre-processing, local maxima, post-processing...). The
pattern_detector function orchestrate all those different steps."""

from __future__ import absolute_import
import numpy as np
import pandas as pd
import scipy.sparse as sp
import pathlib
import warnings
import chromosight.utils.preprocessing as preproc
import chromosight.utils.stats as cus

warnings.filterwarnings("ignore")


def validate_patterns(
    coords,
    matrix,
    conv_mat,
    detectable_bins,
    kernel_matrix,
    drop=True,
    zero_tol=0.3,
    missing_tol=0.75,
):
    """
    Given a list of pattern coordinates and a contact map, drop or flag
    patterns in low detectability regions or too close to matrix boundaries.
    Also returns the surrounding window of Hi-C contacts around each detected
    pattern.

    Parameters
    ----------
    coords : numpy.array of ints
        Coordinates of all detected patterns in the sub matrix. One pattern
        per row, the first column is the matrix row, second column is the
        matrix col.
    matrix : scipy.sparse.csr_matrix
        Hi-C contact map of the sub matrix.
    conv_mat : scipy.sparse.csr_matrix
        Convolution product of the kernel with the Hi-C sub matrix.
    detectable_bins : list of numpy.array
        List of two 1D numpy arrays of ints representing ids of detectable
        rows and columns, respectively.
    kernel_matrix : numpy.array of float
        The kernel that was used for pattern detection on the Hi-C matrix.
    zero_tol : float
        Proportion of zero pixels allowed in a pattern window to
        consider it valid.
    missing_tol : float
        Proportion of missing pixels allowed in a pattern window to
        consider it valid.
    drop : bool
        Whether to discard pattern coordinates and windows from patterns which
        fall outside the matrix or do not pass validation. If those are kept,
        they will be given a score of np.nan and their windows will be filled
        with np.nan.

    Returns
    -------
    filtered_coords : pandas.DataFrame
        Table of coordinates that passed the filters. The dataframe has 3:
        columns: bin1 (rows), bin2 (col) and score (the correlation
        coefficient).
    filtered_windows : numpy.array
        3D numpy array of signal windows around detected patterns. Each window
        spans axes 1 and 2, and they are stacked along axis 0.
    """
    matrix = matrix.tocsr()
    # Pre-compute height, width and half (radius)
    win_h, win_w = kernel_matrix.shape
    half_h, half_w = win_h // 2 + 1, win_w // 2 + 1
    # Store coords to drop
    blacklist = []
    missing_rows = preproc.valid_to_missing(detectable_bins[0], matrix.shape[0])
    missing_cols = preproc.valid_to_missing(detectable_bins[1], matrix.shape[1])

    # Copy coords object and append column for scores
    validated_coords = pd.DataFrame(
        {
            "bin1": coords[:, 0],
            "bin2": coords[:, 1],
            "score": np.full(coords.shape[0], np.nan),
        }
    )
    # validated_coords = np.append(coords, np.zeros((coords.shape[0], 1)), 1)
    # Initialize structure to store pattern windows
    pattern_windows = np.full(
        (coords.shape[0], win_h, win_w), np.nan
    )  # list containing all panel of detected patterns
    for i, l in enumerate(coords):
        p1 = int(l[0])
        p2 = int(l[1])
        high, low = p1 - half_h + 1, p1 + half_h
        left, right = p2 - half_w + 1, p2 + half_w
        # Check for out of bounds errors
        if (
            high >= 0
            and low < matrix.shape[0]
            and left >= 0
            and right < matrix.shape[1]
        ):
            # Subset window from chrom matrix
            pattern_window = matrix[high:low, left:right].toarray()
            n_rows, n_cols = pattern_window.shape
            # Compute positions of missing bins relative to window
            rel_miss_rows = missing_rows - high
            rel_miss_rows = rel_miss_rows[
                (rel_miss_rows >= 0) & (rel_miss_rows < n_rows)
            ]
            rel_miss_cols = missing_cols - left
            rel_miss_cols = rel_miss_cols[
                (rel_miss_cols >= 0) & (rel_miss_cols < n_cols)
            ]
            # Set missing bins to nan
            pattern_window[rel_miss_rows, :] = np.nan
            pattern_window[:, rel_miss_cols] = np.nan

            tot_pixels = n_rows * n_cols
            # Number of uncovered or missing pixels
            tot_zero_pixels = np.sum(pattern_window == 0)
            tot_missing_pixels = np.sum(~np.isfinite(pattern_window))
            # The pattern should not contain more missing pixels that the max
            # value defined in kernel config. This includes both pixels from
            # undetectable bins and zero valued pixels in detectable bins.
            prop_undetected = tot_missing_pixels / tot_pixels
            prop_zero = tot_zero_pixels / (tot_pixels - tot_missing_pixels)
            # Only compute scores for patterns with at least 25% detectable
            # pixels (by default) and less than user-defined proportion of zero
            # pixels
            if (prop_undetected < missing_tol) and (prop_zero < zero_tol):
                validated_coords.score[i] = conv_mat[p1, p2]
                pattern_windows[i, :, :] = pattern_window
            else:
                # Current pattern will be dropped due to undetectable bins
                blacklist.append(i)
        else:
            # Current pattern will be dropped due to out of bound error
            blacklist.append(i)

    # Drop patterns that did not pass filters
    if drop:
        blacklist = np.array(blacklist)
        blacklist_mask = np.zeros(coords.shape[0], dtype=bool)
        if len(blacklist):
            blacklist_mask[blacklist] = True
        filtered_coords = validated_coords.loc[~blacklist_mask, :]
        filtered_windows = pattern_windows[~blacklist_mask, :, :]
    else:
        filtered_coords = validated_coords
        filtered_windows = pattern_windows

    return filtered_coords, filtered_windows


def pileup_patterns(pattern_windows):
    """
    Generate a pileup (arithmetic mean) from a stack of pattern windows.

    Parameters
    ----------
    pattern_windows : numpy.array of floats
        3D numpy array of detected windows. Shape is (N, H, W)
        where N is the number of windows, H the height, and W
        the width of each window.
    Returns
    -------
    numpy.array of floats :
        2D numpy array containing the pileup (arithmetic mean) of
        input windows.
    """
    return np.apply_along_axis(np.nanmean, 0, pattern_windows)


def pattern_detector(
    contact_map,
    kernel_config,
    kernel_matrix,
    coords=None,
    dump=None,
    full=False,
    tsvd=None,
):
    """
    Detect patterns in a contact map by kernel matching, and extract windows
    around the detected patterns. If coordinates are provided, detection is
    skipped and windows are extracted around those coordinates.

    Parameters
    ----------
    contact_map : ContactMap object
        An object containing an inter- or intra-chromosomal Hi-C contact map
        and additional metadata.
    kernel_config : dict
        The kernel configuration, as documented in
        chromosight.utils.io.load_kernel_config
    kernel_matrix : numpy.array
        The kernel matrix to use for convolution as a 2D numpy array
    coords : numpy.array of ints or None
        A table with coordinates of patterns, with one pattern per row
        and 2 columns being the row and column number of the pattern in
        the input contact map. If this is provided, detection is skipped
        and quantification is performed on those coordinates.
    dump : str or None
        Folder in which dumps should be generated after each step of the
        detection process. If None, no dump is generated
    tsvd : float or None
        If a float between 0 and 1 is given, the input kernel is factorised
        using truncated SVD, keeping enough singular vectors to retain this
        proportion of information. Factorisation speeds up convolution at
        the cost of a loss of information. If the number of singular vectors
        required to retain the desired information is disabled by default.

    Returns
    -------
    filtered_chrom_patterns : pandas.DataFrame
        A table of detected patterns with 4 columns: bin1, bin2, score, qvalue.
    chrom_pattern_windows : numpy array
        A 3D array containing the pile of windows around detected patterns.
    """
    km, kn = kernel_matrix.shape
    kh, kw = (km - 1) // 2, (kn - 1) // 2

    def save_dump(base, mat):
        """Define where to save the dump"""
        sp.save_npz(
            pathlib.Path(dump) / f"{contact_map.name}_{base}", mat
        )

    # Define type of analysis.
    run_mode = "detect" if coords is None else "quantify"

    # Do not attempt pattern detection unless matrix is larger than the kernel
    if min(contact_map.matrix.shape) <= max(kernel_matrix.shape):
        return None, None

    # If full is specified, missing bins are accounted for using a mask
    if full:
        missing_mask = preproc.make_missing_mask(
            contact_map.matrix.shape,
            valid_rows=contact_map.detectable_bins[0],
            valid_cols=contact_map.detectable_bins[1],
            max_dist=contact_map.max_dist,
            sym_upper=not contact_map.inter,
        )
    else:
        missing_mask = None

    # Pattern matching operates here
    mat_conv, mat_log10_pvals = normxcorr2(
        contact_map.matrix.tocsr(),
        kernel_matrix,
        max_dist=contact_map.max_dist,
        sym_upper=not contact_map.inter,
        full=full,
        missing_mask=missing_mask,
        tsvd=tsvd,
        pval=True,
        missing_tol=kernel_config["max_perc_undetected"] / 100,
    )
    if dump:
        save_dump("03_normxcorr2", mat_conv)
    # Clean potential missing values
    mat_conv.data[np.isnan(mat_conv.data)] = 0
    # Only keep corrcoefs in scannable range
    if not contact_map.inter:
        mat_conv = preproc.diag_trim(mat_conv.tocsr(), contact_map.max_dist)
        if dump:
            save_dump("04_diag_trim", mat_conv)
    mat_conv = mat_conv.tocoo()
    mat_conv.eliminate_zeros()

    # Only attempt detection if no input coordinates were given
    if run_mode == "detect":
        # Find foci of highly correlated pixels and pick local maxima
        # coords, foci_mat = picker(np.abs(mat_log10_pvals), 5)
        coords, foci_mat = picker(mat_conv, kernel_config["pearson"],)
        # If nothing was detected, no point in resuming
        if coords is None:
            return None, None
        if dump:
            save_dump("05_foci", foci_mat)
    mat = contact_map.matrix.copy()
    det = [d.copy() for d in contact_map.detectable_bins]
    # Zero pad contact and convolution maps and shift missing bins and detected
    # pattern coords before validation if in full mode

    if full:
        mat = mat.tocoo()
        mat = preproc.zero_pad_sparse(mat, kh, kw, fmt="csr")
        mat_conv = preproc.zero_pad_sparse(mat_conv, kh, kw, fmt="csr")
        det[0] += kh
        det[1] += kw
        coords[:, 0] += kh
        coords[:, 1] += kw

    if not contact_map.inter:
        # set the first kh / 2 diagonals in the lower triangle to NaN
        # so that pileups do not count them
        big_k = max(km, kn)
        mat = mat.tocsr()
        mat += sp.diags(
            np.full(big_k, np.nan),
            -np.arange(1, big_k + 1),
            shape=mat.shape,
            format="csr",
        )
        # When detecting 1D pattern, enforce coordinates on diagonal
        # coordinates can be shifted by 1 since we keep the two first
        # diagonals to allow formation of foci via 4-way adjacency
        if kernel_config["max_dist"] == 0:
            coords[:, 0] = coords[:, 1]

    # Extract windows around coordinates and assign a correlation
    # to each pattern. In detection mode, we drop invalid patterns
    # in quantification mode, all input patterns are returned.
    filtered_coords, filtered_windows = validate_patterns(
        coords,
        mat,
        mat_conv.tocsr(),
        det,
        kernel_matrix,
        zero_tol=kernel_config["max_perc_zero"] / 100,
        missing_tol=kernel_config["max_perc_undetected"] / 100,
        drop=True if run_mode == "detect" else False,
    )

    # Shift coordinates of detected patterns back if padding was added
    if full:
        filtered_coords.bin1 -= kh
        filtered_coords.bin2 -= kw

    try:
        filtered_coords["pvalue"] = mat_log10_pvals[
            filtered_coords.bin1, filtered_coords.bin2
        ].A1
    # No coordinate passed the validation filters
    except AttributeError:
        filtered_coords["pvalue"] = None
    # Remove log10 transform and correct p-values for multiple testing
    filtered_coords["pvalue"] = 10 ** filtered_coords["pvalue"]
    return filtered_coords, filtered_windows


def remove_neighbours(patterns, win_size=8):
    """
    Identify patterns that are too close from each other to exclude them.
    The pattern with the highest score are kept in priority.

    Parameters
    ----------
    patterns : numpy.array of float
        2D Array of patterns, with 3 columns: bin1, bin2 and score.
    win_size : int
        The maximum number of pixels at which patterns are considered
        overlapping.

    Returns
    -------
    numpy.array of bool :
        Boolean array indicating which patterns are valid (True values) and
        which are overlapping neighbours (False values)
    """
    # Sort patterns by scores
    sorted_patterns = patterns.copy().sort_values("score", ascending=False)
    # Keep track of patterns to drop
    blacklist = set()
    for i, p in sorted_patterns.iterrows():
        if i not in blacklist:
            close_patterns = np.flatnonzero(
                (np.abs(sorted_patterns.bin1 - p.bin1) < win_size)
                & (np.abs(sorted_patterns.bin2 - p.bin2) < win_size)
            )
            close_patterns_idx = sorted_patterns.index.values[close_patterns]
            close_patterns_idx = close_patterns_idx[close_patterns_idx != i]
            for idx in close_patterns_idx:
                blacklist.add(idx)
    # Get indices of valid patterns as a boolean mask
    whitelist_mask = np.ones(sorted_patterns.shape[0], dtype=bool)
    whitelist_mask[list(blacklist)] = False
    return whitelist_mask


def picker(mat_conv, pearson):
    """
    Pick coordinates of local maxima in a sparse 2D convolution heatmap. A
    threshold computed based on the pearson argument is applied to the heatmap.
    All values below that threshold are set to 0. The coordinate of the maximum
    value in each focus (contiguous region of high values) is returned.

    Parameters
    ----------
    mat_conv : scipy.sparse.coo_matrix of floats
        A 2D sparse convolution heatmap.
    pearson : float
        Minimum correlation coefficient required to consider a pixel as
        candidate. Increasing this value reduces the amount of false
        positive patterns.
    Returns
    -------
    foci_coords : numpy.array of ints
        2D array of coordinates for identified patterns corresponding to
        foci maxima. None is no pattern was detected.
    labelled_mat : scipy.sparse.coo_matrix
        The map of detected foci. Pixels which were assigned to a focus are
        given an integer as their focus ID. Pixels not assigned to a focus
        are set to 0.
    """

    # Filter and binarize sparse matrix based on input threshold
    candidate_mat = mat_conv.copy()
    candidate_mat = candidate_mat.tocoo()
    candidate_mat.data[candidate_mat.data < pearson] = 0
    candidate_mat.data[candidate_mat.data != 0] = 1
    candidate_mat.eliminate_zeros()

    # Check if at least one candidate pixel was found
    if len(candidate_mat.data) > 0:
        # Assign foci identifiers to pixels above threshold
        num_foci, labelled_mat = label_foci(candidate_mat)
        # Remove foci which are too small
        num_foci, labelled_mat = filter_foci(labelled_mat)
        if num_foci == 0:
            return None, None
        mat_conv = mat_conv.tocsr()
        # Will hold the coordinates of the best pixel for each focus
        foci_coords = np.zeros([num_foci, 2], int)
        # Iterate over candidate foci
        # NOTE: There can be jumps between foci id due to post labelling
        # removal of single pixel foci in label_connnected_pixels_sparse.
        # This is why we use focus_rank
        for focus_rank, focus_id in enumerate(np.unique(labelled_mat.data)):
            # Remember 1D indices of datapoint in focus
            focus_idx = np.flatnonzero(labelled_mat.data == focus_id)
            focus_rows, focus_cols = (
                labelled_mat.row[focus_idx],
                labelled_mat.col[focus_idx],
            )
            # Find index of max value within those indices
            focus_pixel_idx = np.argmax(mat_conv[focus_rows, focus_cols])
            # Retrieve row of original index
            original_pixel_idx = focus_idx[focus_pixel_idx]
            focus_pixel_row = labelled_mat.row[original_pixel_idx]
            focus_pixel_col = labelled_mat.col[original_pixel_idx]
            # Save coords of best pixels for this focus in a table
            foci_coords[focus_rank, 0] = focus_pixel_row
            foci_coords[focus_rank, 1] = focus_pixel_col
    else:
        return None, None
    return foci_coords, labelled_mat


def label_foci(matrix):
    """
    Given a sparse matrix of 1 and 0 values, find
    all foci of continuously neighbouring positive pixels
    and assign them a label according to their focus. Horizontal
    and vertical (4-way) adjacency is considered.

    Parameters
    ----------
    matrix : scipy.sparse.coo_matrix of ints
        The input matrix where to label foci. Should be filled with 1
        and 0s.

    Returns
    -------
    num_foci : int
        Number of individual foci identified.
    foci_mat : scipy.sparse.coo_matrix:
        The matrix with values replaced by their respective foci
        labels.

    Example
    -------
    >>> M.todense()
    array([[1 0 0 0]
           [1 0 1 0]
           [1 0 1 1]
           [0 0 0 0]])
    >>> num_foci, foci_mat = label_foci(M)
    >>> num_foci
    2
    >>>foci_mat.todense()
    array([[1 0 0 0]
           [1 0 2 0]
           [1 0 2 2]
           [0 0 0 0]])
    """
    # step 1 : ensure that coo format is double ranked (row, col).
    matrix_sp = sp.coo_matrix(sp.csr_matrix(matrix))

    nb_row, nb_col = matrix_sp.shape
    nb_nodes = matrix_sp.nnz

    # step 2: prepare structured array 'coo' for later double sorting
    dtype = [("row", int), ("col", int)]
    coo = np.zeros(matrix_sp.nnz, dtype=dtype)
    coo["row"] = matrix_sp.row
    coo["col"] = matrix_sp.col

    # step 3: find neigbors to the right
    coo_rc = np.argsort(coo, order=["row", "col"])
    row_rc = coo["row"][coo_rc]
    col_rc = coo["col"][coo_rc]

    # sanity check
    if not (np.all(coo_rc == np.arange(len(coo_rc)))):
        raise ValueError("matrix_sp is not properly double sorted")

    # Compute row and col indices shift between right-neighbors
    diff_row_rc = row_rc[1:] - row_rc[:-1]
    diff_col_rc = col_rc[1:] - col_rc[:-1]
    # Right connected pixels are on the same row and neighboring cols
    right_connected_p = (diff_row_rc == 0) & (diff_col_rc == 1)
    right_connected_k = np.flatnonzero(right_connected_p)

    right_node1 = right_connected_k
    right_node2 = right_connected_k + 1

    # step 4: find neigbors beneath
    coo_cr = np.argsort(coo, order=["col", "row"])
    row_cr = coo["row"][coo_cr]
    col_cr = coo["col"][coo_cr]

    diff_row_cr = row_cr[1:] - row_cr[:-1]
    diff_col_cr = col_cr[1:] - col_cr[:-1]

    lower_connected_p = (diff_row_cr == 1) & (diff_col_cr == 0)
    lower_connected_k = np.flatnonzero(lower_connected_p)

    cr2rc = np.arange(len(coo_cr))[coo_cr]
    lower_node1 = cr2rc[lower_connected_k]
    lower_node2 = cr2rc[lower_connected_k + 1]

    # step 5: build adjacency matrix
    node1 = np.concatenate([right_node1, lower_node1])
    node2 = np.concatenate([right_node2, lower_node2])
    data = np.ones(len(node1), int)
    adj_mat = sp.coo_matrix((data, (node1, node2)), shape=(nb_nodes, nb_nodes))

    # step 6: find connected components
    num_foci, foci = sp.csgraph.connected_components(adj_mat, directed=False)
    foci_mat = sp.coo_matrix(
        (foci + 1, (matrix_sp.row, matrix_sp.col)), shape=(nb_row, nb_col)
    )

    return num_foci, foci_mat


def filter_foci(foci_mat, min_size=2):
    """
    Given an input matrix of labelled foci (continuous islands of equal nonzero
    values), filters out foci consisting of fewer pixels than min_size.

    Parameters
    ----------
    foci_mat : scipy.sparse.coo_matrix
        Input matrix of labelled foci. Pixels are numbered according to their
        respective foci. Pixels that are not assigned to a focus are 0.
    min_size : int
        Minimum number of pixels required to keep a focus. Pixels belonging to
        smaller foci will be set to 0.

    Returns
    -------
    num_filtered : int
        Number of foci remaining after filtering.
    filtered_mat : scipy.sparse.coo_matrix
        Matrix of filtered foci.
    """
    foci_data = foci_mat.data
    focus_id, focus_size = np.unique(foci_data, return_counts=True)
    # Remove foci smaller than min_size
    small_foci = focus_id[focus_size < min_size]
    # mask small foci for removal
    for focus_id in small_foci:
        foci_data[foci_data == focus_id] = 0
    # Copy input matrix and replace data with filtered foci
    filtered_mat = foci_mat.copy()
    filtered_mat.data = foci_data
    # Force updating to make explicit zeros implicit
    filtered_mat.eliminate_zeros()
    num_filtered = len(focus_size[focus_size >= min_size])

    return num_filtered, filtered_mat


def xcorr2(signal, kernel, threshold=1e-4, tsvd=None):
    """
    Cross correlate a dense or sparse 2D signal with a dense 2D kernel.

    Parameters
    ----------
    signal: scipy.sparse.csr_matrix or numpy.array of floats
        A 2-dimensional numpy array Ms x Ns acting as the detrended Hi-C map.
    kernel: numpy.array of floats
        A 2-dimensional numpy array Mk x Nk acting as the pattern template.
    threshold : float
        Convolution score below which pixels will be set back to zero to save
        on time and memory.
    tsvd : float or None
        If a float between 0 and 1 is given, the input kernel is factorised
        using truncated SVD, keeping enough singular vectors to retain this
        proportion of information. Factorisation speeds up convolution at
        the cost of a loss of information. If the number of singular vectors
        required to retain the desired information isDisabled by default.
    -------
    out: scipy.sparse.csr_matrix or numpy.array
        Convolution product of signal by kernel. Same type as the input signal.
    """
    if tsvd is not None:
        kernel = preproc.factorise_kernel(kernel, prop_info=tsvd)
    if sp.issparse(signal):
        conv = _xcorr2_sparse(signal.tocsr(), kernel, threshold=threshold)
    else:
        conv = _xcorr2_dense(signal, kernel, threshold=threshold)
    return conv


def _xcorr2_sparse(signal, kernel, threshold=1e-6):
    """
    Cross correlate a sparse 2D signal with a dense 2D kernel.

    Parameters
    ----------
    signal: scipy.sparse.csr_matrix
        A 2-dimensional numpy array Ms x Ns acting as the detrended Hi-C map.
    kernel: numpy.array of floats or tuple of numpy.arrays
        A 2-dimensional numpy array Mk x Nk acting as the pattern template. Can
        also be a factorised kernel.
    threshold : float
        Convolution score below which pixels will be set back to zero to save
        on time and memory.
    Returns
    -------
    out: scipy.sparse.csr_matrix
        Convolution product of signal by kernel.
    """
    sm, sn = signal.shape

    if type(kernel) is tuple:
        kernel_l, kernel_r = kernel
        km = kernel_l.shape[0]
        kn = kernel_r.shape[1]
        if kernel_l.shape[1] != kernel_r.shape[0]:
            raise ValueError("Kernel factorisation is invalid")
        n_factors = kernel_l.shape[1]
        for f in range(n_factors):
            subkernel_l = sp.diags(
                kernel_l[:, f], np.arange(km), shape=(sm - km + 1, sm), format="dia",
            )
            subkernel_r = sp.diags(
                kernel_r[f, :], -np.arange(kn), shape=(sn, sn - kn + 1), format="dia",
            )
            if f == 0:
                out = (subkernel_l @ signal) @ subkernel_r
            else:
                out += (subkernel_l @ signal) @ subkernel_r
    else:
        km, kn = kernel.shape

        # Sanity checks
        if sp.issparse(kernel):
            raise ValueError("cannot handle kernel in sparse format")
        if not sp.issparse(signal):
            raise ValueError("cannot handle signal in dense format")
        # Check of kernel is constant (uniform)
        constant_kernel = np.nan
        if np.allclose(kernel, np.tile(kernel[0, 0], kernel.shape), rtol=1e-08):
            constant_kernel = kernel[0, 0]

        out = sp.csc_matrix((sm - km + 1, sn - kn + 1), dtype=np.float64)

        # Simplified convolution for the special case where kernel is constant:
        if np.isfinite(constant_kernel):
            l_subkernel_sp = sp.diags(
                constant_kernel * np.ones(km),
                np.arange(km),
                shape=(sm - km + 1, sm),
                format="dia",
            )
            r_subkernel_sp = sp.diags(
                np.ones(kn), -np.arange(kn), shape=(sn, sn - kn + 1), format="dia",
            )
            out = (l_subkernel_sp @ signal) @ r_subkernel_sp
        # convolution code for general case
        # 1. 2D kernel composed of 1D filters, each col being a 1D filter.
        # 2. input remains unchanged, and each 1D kernel is unrolled into
        # a sparse toeplitz matrix.
        # 3. Each 1D conv is computed via a sparse matrix x vector product.
        # It is fast because the product is delegated to numpy.
        else:
            for kj in range(kn):
                subkernel_sp = sp.diags(
                    kernel[:, kj], np.arange(km), shape=(sm - km + 1, sm), format="csr",
                )
                out += subkernel_sp.dot(signal[:, kj : sn - kn + 1 + kj])

    # Set very low pixels to 0
    out.data[np.abs(out.data) < threshold] = 0
    out.eliminate_zeros()

    # Resize matrix to original dimensions
    out = preproc.zero_pad_sparse(
        out, margin_h=(kn - 1) // 2, margin_v=(km - 1) // 2, fmt="csr"
    )
    return out


def _xcorr2_dense(signal, kernel, threshold=1e-6):
    """Cross correlate a dense 2D signal with a dense 2D kernel.

    Parameters
    ----------
    signal: numpy.array
        A 2-dimensional numpy array Ms x Ns acting as the detrended Hi-C map.
    kernel: array_like
        A 2-dimensional numpy array Mk x Nk acting as the pattern template.
    threshold : float
        Convolution score below which pixels will be set back to zero to save
        on time and memory.
    Returns
    -------
    out: numpy.array
        Convolution product of signal by kernel.
    """
    sm, sn = signal.shape
    if type(kernel) is tuple:
        kernel_l, kernel_r = kernel
        km = kernel_l.shape[0]
        kn = kernel_r.shape[1]
        if kernel_l.shape[1] != kernel_r.shape[0]:
            raise ValueError("Kernel factorisation is invalid")
        n_factors = kernel_l.shape[1]
        for f in range(n_factors):
            subkernel_l = sp.diags(
                kernel_l[:, f], np.arange(km), shape=(sm - km + 1, sm), format="dia",
            )
            subkernel_r = sp.diags(
                kernel_r[f, :], -np.arange(kn), shape=(sn, sn - kn + 1), format="dia",
            )
            if f == 0:
                out = (subkernel_l @ signal) @ subkernel_r
            else:
                out += (subkernel_l @ signal) @ subkernel_r
    else:
        km, kn = kernel.shape
        # Kernel (half) height and width
        constant_kernel = np.nan
        if np.allclose(kernel, np.tile(kernel[0, 0], kernel.shape), rtol=1e-08):
            constant_kernel = kernel[0, 0]
        out_wo_margin = np.zeros([sm - km + 1, sn - kn + 1])
        # Simplified convolution for the special case where kernel is constant:
        if np.isfinite(constant_kernel):
            l_subkernel_sp = sp.diags(
                constant_kernel * np.ones(km),
                np.arange(km),
                shape=(sm - km + 1, sm),
                format="dia",
            )
            r_subkernel_sp = sp.diags(
                np.ones(kn), -np.arange(kn), shape=(sn, sn - kn + 1), format="dia",
            )
            out_wo_margin = (l_subkernel_sp @ signal) @ r_subkernel_sp
        # convolution code for general case
        # 1. 2D kernel composed of 1D filters, each col being a 1D filter.
        # 2. input remains unchanged, and each 1D kernel is unrolled into
        # a sparse toeplitz matrix.
        # 3. Each 1D conv is computed via a sparse matrix x vector product.
        # It is fast because the product is delegated to numpy.
        else:
            for kj in range(kn):
                # Make a toeplitz matrix from the kj'th column of the kernel
                subkernel_sp = sp.diags(
                    kernel[:, kj], np.arange(km), shape=(sm - km + 1, sm), format="csr",
                )
                # Note: We compute the dot product of the subkernel with the signal 
                # shifted by kj.
                out_wo_margin += subkernel_sp @ signal[:, kj : sn - kn + 1 + kj]

    kh = (km - 1) // 2
    kw = (kn - 1) // 2
    # Add margins of zeros where kernel overlaps edges
    out = np.zeros([sm, sn])
    out[kh:-kh, kw:-kw] = out_wo_margin
    # Set very low pixels to 0
    out[np.abs(out) < threshold] = 0
    return out


def normxcorr2(
    signal,
    kernel,
    max_dist=None,
    sym_upper=False,
    full=False,
    missing_mask=None,
    missing_tol=0.75,
    tsvd=None,
    pval=False,
):
    """
    Computes the normalized cross-correlation of a dense or sparse signal and a
    dense kernel. The resulting matrix contains Pearson correlation
    coefficients.

    Parameters
    ----------
    signal : scipy.sparse.csr_matrix or numpy.array
        The input processed Hi-C matrix.
    kernel : numpy.array
        The pattern kernel to use for convolution.
    max_dist : int
        Maximum scan distance, in number of bins from the diagonal. If None,
        the whole matrix is convoluted. Otherwise, pixels further than this
        distance from the diagonal are set to 0 and ignored for performance.
        Only useful for intrachromosomal matrices.
    sym_upper : False
        Whether the matrix is symmetric and upper triangle. True for
        intrachromosomal
        matrices.
    missing_mask : scipy.sparse.coo_matrix of ints
        Matrix defining which pixels are missing (1) or not (0).
    full : bool
        If True, convolutions will be made in 'full' mode; the matrix is first
        padded with margins to allow scanning to the edges, and missing bins
        are also masked to exclude them when computing correlation scores.
        Computationally intensive
    missing_mask : scipy.sparse.csr_matrix of bool or None
        Mask matrix denoting missing bins, where missing is denoted as True and
        valid as False. Can be None to ignore missing bin information. Only
        taken into account when full=True.
    missing_tol : float
        Proportion of missing values allowed in windows to keep the correlation
        coefficients.
    tsvd : float or None
        If a float between 0 and 1 is given, the input kernel is factorised
        using truncated SVD, keeping enough singular vectors to retain this
        proportion of information. Factorisation speeds up convolution at
        the cost of a loss of information. If the number of singular vectors
        required to retain the desired information isDisabled by default.
    pval : bool
        Whether to return a matrix of p-values.

    Returns
    -------
    scipy.sparse.csr_matrix or numpy.array
        The sparse matrix of correlation coefficients. Same type as the input
        signal.
    scipy.sparse.csr_matrix or numpy.array or None
        A map of Benjamini-Hochberg corrected p-values (q-values). Same type as
        the input signal. If pval=False, this will be None.
    """

    if missing_mask is not None:
        if not sp.issparse(missing_mask):
            raise ValueError("Missing mask must be a sparse matrix.")
        if not signal.shape == missing_mask.shape:
            raise ValueError("Signal and missing mask do not have the same shape")
        if missing_mask.dtype != bool:
            raise ValueError(
                f"Missing mask dtype is {missing_mask.dtype}. Should be bool."
            )

        if min(kernel.shape) >= max(signal.shape):
            raise ValueError("cannot have kernel bigger than signal")
        preproc.check_missing_mask(signal, missing_mask)

    if sp.issparse(kernel):
        raise ValueError("cannot handle kernel in sparse format")

    if not (kernel.std() > 0):
        raise ValueError("Cannot have flat kernel.")
    if sp.issparse(signal):
        corr, pvals = _normxcorr2_sparse(
            signal,
            kernel,
            max_dist=max_dist,
            sym_upper=sym_upper,
            full=full,
            missing_mask=missing_mask,
            missing_tol=missing_tol,
            tsvd=tsvd,
            pval=pval,
        )
    else:
        corr, pvals = _normxcorr2_dense(
            signal,
            kernel,
            max_dist=max_dist,
            sym_upper=sym_upper,
            full=full,
            missing_tol=missing_tol,
            missing_mask=missing_mask,
            tsvd=tsvd,
            pval=pval,
        )
    return corr, pvals


def _normxcorr2_sparse(
    signal,
    kernel,
    max_dist=None,
    sym_upper=False,
    full=False,
    missing_mask=None,
    missing_tol=0.75,
    tsvd=None,
    pval=False,
):
    """Computes the normalized cross-correlation of a sparse signal and a
    dense kernel. The resulting sparse matrix contains Pearson correlation
    coefficients.

    Parameters
    ----------
    signal : scipy.sparse.csr_matrix
        The input processed Hi-C matrix.
    kernel : numpy.array
        The pattern kernel to use for convolution. Can be a factorised kernel
        stored in a tuple.
    max_dist : int
        Maximum scan distance, in number of bins from the diagonal. If None,
        the whole matrix is convoluted. Otherwise, pixels further than this
        distance from the diagonal are set to 0 and ignored for performance.
        Only useful for intrachromosomal matrices.
    sym_upper : False
        Whether the matrix is symmetric and upper triangle. True for
        intrachromosomal matrices.
    missing_mask : scipy.sparse.csr_matrix of bools
        Matrix defining which pixels are missing (True) or not (False).
    missing_tol : float
        Proportion of missing values allowed in windows to keep the correlation
        coefficients.
    full : bool
        Whether to run in 'full' mode, which means enclosing the signal in an
        exterior frame and computing the correlation up to the edges.
    tsvd : float or None
        If a float between 0 and 1 is given, the input kernel is factorised
        using truncated SVD, keeping enough singular vectors to retain this
        proportion of information. Factorisation speeds up convolution at
        the cost of a loss of information. If the number of singular vectors
        required to retain the desired information isDisabled by default.
    pval : bool
        Whether to return a matrix of p-values.

    Returns
    -------
    scipy.sparse.csr_matrix
        The sparse matrix of correlation coefficients
    """
    mk, nk = kernel.shape
    ms, ns = signal.shape

    # Generate constant kernel
    kernel1 = np.ones(kernel.shape)
    kernel_size = mk * nk

    # In full mode, we compute the convolution with all pixels in the input
    # signal. We need to add a margin around the input to allow the kernel to
    # be centered on the edges.
    if full:
        # Create a vertical margin and use it to pad the signal
        tmp = sp.csr_matrix((mk - 1, ns))
        framed_sig = sp.vstack([tmp, signal, tmp], format=signal.format)
        # Same for the horizontal margin
        tmp = sp.csr_matrix((ms + 2 * (mk - 1), nk - 1))
        framed_sig = sp.hstack([tmp, framed_sig, tmp], format=signal.format)
        # If a missing mask was specified, use it
        if missing_mask is not None:
            # Add margins around missing mask.
            framed_missing_mask = preproc.frame_missing_mask(
                missing_mask, kernel.shape, sym_upper=sym_upper, max_dist=max_dist,
            )
        else:
            framed_missing_mask = None
    else:
        framed_sig = signal.copy()
        if missing_mask is None:
            framed_missing_mask = None
        else:
            framed_missing_mask = missing_mask.copy()
    # Ignore missing values
    if framed_missing_mask is None:
        kernel_mean = float(kernel.mean())
        kernel_std = float(kernel.std())
        # out contains framed_sig mean
        out = xcorr2(framed_sig, kernel1 / kernel_size)

        denom = xcorr2(framed_sig.power(2), kernel1 / kernel_size) - out.power(2)
        denom = denom.sqrt() * kernel_std
        # quick and dirty hack to robustify: numerical-zeros are turned into
        # real-zeros
        denom.data[np.abs(denom.data) < 1e-10] = 0.0
        denom.data = 1.0 / denom.data

        # store numerator directly in 'out' to avoid multiple replication of
        # data
        out = xcorr2(framed_sig, kernel / kernel_size, tsvd=tsvd) - out * kernel_mean
        # Multiply by invert since sparse division is not allowed
        out = out.multiply(denom)

    else:
        # Safety check to make sure mask matches signal
        preproc.check_missing_mask(framed_sig, framed_missing_mask)
        kernel_sum = np.sum(kernel)
        kernel_mean = kernel_sum / kernel_size
        kernel2_sum = np.sum(kernel ** 2)
        kernel2_mean = kernel2_sum / kernel_size
        # Compute convolution of uniform kernel with missing mask to get number
        # of missing pixels in each window. Will be plugged into Pearson at
        # the end.
        ker1_coo = xcorr2(framed_missing_mask, kernel1).tocoo()
        # From now on, ker1_coo.data contains the number of 'present' samples.
        # (where there is at least one missing pixel)
        ker1_coo.data = kernel_size - ker1_coo.data
        # Compute mean corrected with with number of missing elements (wm)
        kernel_mean_wm = (
            kernel_sum
            - xcorr2(framed_missing_mask, kernel, tsvd=tsvd)[
                ker1_coo.row, ker1_coo.col
            ].A1
        ) / ker1_coo.data
        kernel2_mean_wm = (
            kernel2_sum
            - xcorr2(framed_missing_mask, kernel ** 2, tsvd=tsvd)[
                ker1_coo.row, ker1_coo.col
            ].A1
        ) / ker1_coo.data

        # store signal mean directly in 'out' to avoid multiple replication of
        # data
        out = xcorr2(framed_sig, kernel1 / kernel_size)
        out[ker1_coo.row, ker1_coo.col] = (
            out[ker1_coo.row, ker1_coo.col].A1 * kernel_size / ker1_coo.data
        )

        denom = xcorr2(framed_sig.power(2), kernel1 / kernel_size)
        denom[ker1_coo.row, ker1_coo.col] = (
            denom[ker1_coo.row, ker1_coo.col].A1 * kernel_size / ker1_coo.data
        )

        denom = (denom - out.power(2)) * (kernel2_mean - kernel_mean ** 2)
        denom[ker1_coo.row, ker1_coo.col] = (
            denom[ker1_coo.row, ker1_coo.col].A1
            / (kernel2_mean - kernel_mean ** 2)
            * (kernel2_mean_wm - kernel_mean_wm ** 2)
        )
        denom = denom.sqrt()

        # ensure that enough data points are inside of window
        denom[
            ker1_coo.row[ker1_coo.data < int((1 - missing_tol) * kernel_size)],
            ker1_coo.col[ker1_coo.data < int((1 - missing_tol) * kernel_size)],
        ] = 0.0

        # store numerator directly in 'out' to avoid multiple copies of data
        out *= kernel_mean
        out[ker1_coo.row, ker1_coo.col] = (
            out[ker1_coo.row, ker1_coo.col].A1
            * kernel_mean_wm
            * ker1_coo.data
            / (kernel_mean * kernel_size)
        )
        out = xcorr2(framed_sig, kernel / kernel_size, tsvd=tsvd) - out
        out[ker1_coo.row, ker1_coo.col] = (
            out[ker1_coo.row, ker1_coo.col].A1 * kernel_size / ker1_coo.data
        )

        # Remember where very low values are located in denom to avoid nans
        denom_0 = abs(denom.data) < 1e-10
        # take inverse, because 2 sparse mats can't be divided
        denom.data[~denom_0] = 1 / denom.data[~denom_0]
        denom.data[denom_0] = 0.0
        out = out.multiply(denom)

    # if (max_dist is not None) and sym_upper:
    #    # Trim diagonals further than max_scan_distance
    #    out = preproc.diag_trim(out.tocsr(), max_dist)

    if sym_upper:
        out = sp.triu(out)
    out = out.tocoo()
    out.data[~np.isfinite(out.data)] = 0.0
    # Remove near zero coeffs to spare memory
    out.data[np.abs(out.data) < 0] = 0.0
    # Prevent rounding errors to generate values above 1 or below -1
    out.data[out.data < -1] = -1.0
    out.data[out.data > 1] = 1.0
    out.eliminate_zeros()
    if pval:
        pvals = out.copy()
        if full and framed_missing_mask is not None:
            try:
                # Get number of values for each coeff
                n_obs = ker1_coo.tocsr()[pvals.row, pvals.col].A1
                # Replace implicit n_obs by total kernel size
                n_obs[n_obs == 0] = kernel_size
                pvals.data = cus.corr_to_pval(out.data, n_obs)
            # No nonzero coeff in the matrix, skip calculation of p-values
            except AttributeError:
                pass
        else:
            pvals.data = cus.corr_to_pval(out.data, kernel_size)
        pvals = pvals.tocsr()
        if full:
            pvals = pvals[mk - 1 : -mk + 1, nk - 1 : -nk + 1]
    else:
        pvals = None
    out = out.tocsr()
    if full:
        out = out[mk - 1 : -mk + 1, nk - 1 : -nk + 1]

    return out, pvals


def _normxcorr2_dense(
    signal,
    kernel,
    max_dist=None,
    sym_upper=False,
    full=None,
    missing_mask=None,
    tsvd=None,
    missing_tol=0.75,
    pval=False,
):
    """Computes the normalized cross-correlation of a dense or sparse signal
    and a dense kernel. The resulting matrix contains Pearson correlation
    coefficients.

    Parameters
    ----------
    signal : numpy.array
        The input processed Hi-C matrix.
    kernel : numpy.array or tuple of numpy.arrays
        The pattern kernel to use for convolution. Can be a factorised kernel
        stored in a tuple.
    max_dist : int
        Maximum scan distance, in number of bins from the diagonal. If None,
        the whole matrix is convoluted. Otherwise, pixels further than this
        distance from the diagonal are set to 0 and ignored for performance.
        Only useful for intrachromosomal matrices.
    sym_upper : False
        Whether the matrix is symmetric and upper triangle. True for
        intrachromosomal matrices.
    full : bool
    missing_mask : numpy.array of bools
        Nump
    missing_tol : float
        Proportion of missing values allowed in windows to keep the correlation
        coefficients.
    tsvd : float or None
        If a float between 0 and 1 is given, the input kernel is factorised
        using truncated SVD, keeping enough singular vectors to retain this
        proportion of information. Factorisation speeds up convolution at
        the cost of a loss of information. If the number of singular vectors
        required to retain the desired information isDisabled by default.
    pval : bool
        Whether to return a matrix of p-values.

    Returns
    -------
    numpy.array
        The sparse matrix of correlation coefficients
    """
    mk, nk = kernel.shape
    ms, ns = signal.shape

    # Convert numpy matrices to array to avoid operator overloading
    if isinstance(signal, np.matrix):
        signal = np.array(signal)
    if isinstance(kernel, np.matrix):
        kernel = np.array(kernel)

    kernel_size = mk * nk
    kernel1 = np.ones(kernel.shape)
    if full:
        framed_sig = np.zeros([ms + 2 * (mk - 1), ns + 2 * (nk - 1)])
        framed_sig[mk - 1 : -mk + 1, nk - 1 : -nk + 1] = signal
        if missing_mask is not None:
            # Add margins around missing mask.
            framed_missing_mask = preproc.frame_missing_mask(
                missing_mask, kernel.shape, sym_upper=sym_upper, max_dist=max_dist,
            )
    else:
        framed_sig = signal
        if missing_mask is not None:
            framed_missing_mask = missing_mask.copy()
    # Pearson correlation
    kernel_mean = float(kernel.mean())
    kernel_std = float(kernel.std())

    framed_sig_mean = xcorr2(framed_sig, kernel1 / kernel_size)

    if missing_mask is None:
        out = xcorr2(framed_sig, kernel / kernel_size) - framed_sig_mean * kernel_mean
        denom = kernel_std * np.sqrt(
            xcorr2(framed_sig ** 2, kernel1 / kernel_size) - framed_sig_mean ** 2
        )
        denom_0 = abs(denom) < 1e-10
        out[~denom_0] /= denom[~denom_0]
        out[denom_0] = 0.0

    else:
        kernel_size = mk * nk - xcorr2(framed_missing_mask, kernel1).toarray()
        kernel_mean = (
            np.sum(kernel) - xcorr2(framed_missing_mask, kernel).toarray()
        ) / kernel_size
        # store signal0_mean in 'out' to avoid multiple replication of data
        out = xcorr2(framed_sig, kernel1) / kernel_size
        denom = xcorr2(framed_sig ** 2, kernel1) / kernel_size
        # Use multiplicative coeff to correct denom by number of missing values
        denom = np.sqrt(
            (denom - out ** 2)
            * (
                (
                    np.sum(kernel ** 2)
                    - xcorr2(framed_missing_mask, kernel ** 2).toarray()
                )
                / kernel_size
                - kernel_mean ** 2
            )
        )
        # ensure that enough data points are inside of window
        denom[kernel_size < int((1 - missing_tol) * mk * nk)] = 0.0

        # store numerator directly in 'out' to avoid multiple copies of data
        out = xcorr2(framed_sig, kernel / kernel_size) - out * kernel_mean
        # Remember where very low values are located in denom to avoid nans
        denom_0 = abs(denom) < 1e-10

        out[~denom_0] /= denom[~denom_0]
        out[denom_0] = 0.0

    if sym_upper:
        out = np.triu(out)
    out[~np.isfinite(out)] = 0.0
    out[np.abs(out) < 0] = 0.0
    # Prevent rounding errors to generate values above 1 or below -1
    out[out < -1] = -1.0
    out[out > 1] = 1.0
    if pval:
        if full:
            # Get number of values for each coeff
            n_obs = kernel_size.flatten()
        else:
            n_obs = kernel_size
        pvals = cus.corr_to_pval(out.flatten(), n_obs).reshape(out.shape)
        if full:
            pvals = pvals[mk - 1 : -mk + 1, nk - 1 : -nk + 1]
    else:
        pvals = None
    if full:
        out = out[mk - 1 : -mk + 1, nk - 1 : -nk + 1]
    return out, pvals



#!/usr/bin/env python3
# coding: utf-8

from .version import __version__ as version

__author__ = (
    "Cyril Matthey-Doret, Lyam Baudry, Axel Breuer, Remi Montagne, ",
    "Nadège Guiglielmoni, Vittore Scolari, Etienne Jean, Arnaud Campeas, ",
    "Philippe-Henri Chanut, Edgar Oriol, Adrien Meot, Laurent Politis, ",
    "Antoine Vigouroux, Pierrick Moreau, Romain Koszul, Axel Cournac",
)
__copyright__ = "Copyright © 2017-2018, Institut Pasteur, Paris, France"
__credits__ = ["Hackathon ENGIE/Institut Pasteur teams"]
__license__ = "GPLv3"
__maintainer__ = "Axel Cournac, Cyril Matthey-Doret, Nadège Guigliemoni"
__email__ = "axel.cournac@pasteur.fr"
__status__ = "Beta"
__version__ = version



"""
Chromosight's kernel submodule contains each default kernel in the form of a
dictionary whith the kernel name. The items of the dictionaries are the
key-value pairs from the kernel's json file, with the kernel matrices
pre-loaded under the "kernels" key. Here is the kernel submodule can be used to
extract the first borders kernel:

.. code-block:: python

    import chromosight.kernels as ck
    kernel = ck.borders['kernels'][0]

A list of all available kernel names can also be accessed directly:

.. code-block:: python

    import chromosight.kernels as ck
    names = ck.kernel_names

"""
from chromosight.utils.io import load_kernel_config
import pathlib
import sys

# Here, each pattern config file detected in the kernels directory is loaded
# and made available as a variable in the API

# Get parent module (chromosight.kernels)
current_module = sys.modules[__name__]
# Find all json files in kernel dir
kernel_dir = pathlib.Path(__file__).parents[0]
kernel_names = []
for kernel_file in kernel_dir.glob("*.json"):
    # Get pattern names based on config file name
    pattern_name = str(kernel_file.with_suffix("").name)
    # Declare pattern configs as module level (chromosight.kernels) variables
    setattr(
        current_module,
        pattern_name,
        load_kernel_config(pattern_name, custom=False),
    )
    kernel_names.append(pattern_name)

setattr(current_module, 'kernel_names', kernel_names)



#!/usr/bin/env python3
# coding: utf-8

"""Chromosight's preprocessing submodule implements a number of functions to
operate on Hi-C contact maps before detection. These functions can be used to
improve the signal or filter unneeded signal. There are also functions to edit
(zoom, crop, factorize) kernel matrices.
"""

import sys
import numpy as np
import numpy.linalg as la
import scipy.stats as ss
import scipy.sparse as sp
import scipy.ndimage as ndi
from sklearn.isotonic import IsotonicRegression


def erase_missing(signal, valid_rows, valid_cols, sym_upper=True):
    """
    Given a sparse matrix, set all pixels in missing (invalid) bins to 0.

    Parameters
    ----------
    signal : scipy.sparse.csr_matrix of floats
        Input signal on which to erase values.
    valid_rows : numpy.ndarray of ints
        Indices of rows considered valid (not missing).
    valid_cols : numpy.ndarray of ints
        Indices of columns considered valid (not missing).
    sym_upper : bool
        Define if the input signal is upper symmetric.

    Returns
    -------
    scipy.sparse.csr_matrix
        The input signal with all values in missing bins set to 0
    """
    if sym_upper and sp.issparse(signal):
        if np.any(valid_rows != valid_cols):
            raise ValueError(
                "Valid rows and columns must be identical with sym_upper=True"
            )
        if signal.shape[0] != signal.shape[1]:
            raise ValueError(
                "Input matrix must be square when using sym_upper=True"
            )
        # Make a boolean mask from good bins
        good_mask = np.isin(range(signal.shape[0]), valid_rows)
        # Set all pixels in a nondetectable bin to 0
        # For faster masking of bins, mask bins using dot product with an
        # identify matrix where bad bins have been masked on the diagonal
        # E.g. if removing the second bin (row and column):
        # 1 0 0     9 6 5     1 0 0     9 0 5
        # 0 0 0  X  6 8 7  X  0 0 0  =  0 0 0
        # 0 0 1     6 7 8     0 0 1     6 0 8
        mask_mat = sp.eye(signal.shape[0])
        mask_mat.data[0][~good_mask] = 0
        erased = mask_mat.dot(signal).dot(mask_mat)
    else:
        # Get a boolean array of missing (1) and valid (0) rows
        missing_rows = valid_to_missing(valid_rows, signal.shape[0])
        missing_cols = valid_to_missing(valid_cols, signal.shape[1])
        erased = signal.copy()
        erased[missing_rows, :] = 0
        erased[:, missing_cols] = 0

    return erased


def set_mat_diag(mat, diag=0, val=0):
    """
    Set the nth diagonal of a symmetric 2D numpy array to a fixed value.
    Operates in place.

    Parameters
    ----------
    mat : numpy.ndarray
        Symmetric 2D array of floats.
    diag : int
        0-based index of the diagonal to modify. Use negative values for the
        lower half.
    val : float
        Value to use for filling the diagonal
    """
    m = mat.shape[0]
    step = m + 1
    start = diag
    end = m ** 2 - diag * m
    mat.flat[start:end:step] = val


def diag_trim(mat, n):
    """
    Trim an upper triangle sparse matrix so that only the first n diagonals
    are kept.

    Parameters
    ----------

    mat : scipy.sparse.csr_matrix or numpy.ndarray
        The sparse matrix to be trimmed
    n : int
        The number of diagonals from the center to keep (0-based).

    Returns
    -------
    scipy.sparse.dia_matrix or numpy.ndarray:
        The diagonally trimmed upper triangle matrix with only the first
        n diagonal.
    """
    if sp.issparse(mat):
        if mat.format != "csr":
            raise ValueError("input type must be scipy.sparse.csr_matrix")
        # Trim diagonals by removing all elements further than n in the
        # upper triangle
        trimmed = sp.tril(mat, n, format="csr")
        trimmed = sp.triu(trimmed, format="csr")
    else:
        trimmed = mat.copy()
        n_diags = trimmed.shape[0]
        for diag in range(n, n_diags):
            set_mat_diag(trimmed, diag, 0)
        return trimmed

    return trimmed


def distance_law(
    matrix, detectable_bins=None, max_dist=None, smooth=True, fun=np.nanmean
):
    """
    Computes genomic distance law by averaging over each diagonal in the upper
    triangle matrix. If a list of detectable bins is provided, pixels in
    missing bins will be excluded from the averages. A maximum distance can be
    specified to define how many diagonals should be computed.

    parameters
    ----------
    matrix: scipy.sparse.csr_matrix
        the input matrix to compute distance law from.
    detectable_bins : numpy.ndarray of ints
        An array of detectable bins indices to consider when computing
        distance law.
    max_dist : int
        Maximum distance from diagonal, in number of bins in which to compute
        distance law
    smooth : bool
        Whether to use isotonic regression to smooth the distance law.
    fun : callable
        A function to apply on each diagonal. Defaults to mean.

    Returns
    -------
    dist: np.ndarray
        the output genomic distance law.

    example
    -------
        >>> m = np.ones((3,3))
        >>> m += np.array([1,2,3])
        >>> m
        array([[2., 3., 4.],
               [2., 3., 4.],
               [2., 3., 4.]])
        >>> distance_law(csr_matrix(m))
        array([3. , 3.5, 4. ])

    """
    mat_n = matrix.shape[0]
    if max_dist is None:
        max_dist = mat_n
    n_diags = min(mat_n, max_dist + 1)
    dist = np.zeros(mat_n)
    if detectable_bins is None:
        detectable_bins = np.array(range(mat_n))

    for diag in range(n_diags):
        # Find detectable which fall in diagonal
        detect_mask = np.zeros(mat_n, dtype=bool)
        detect_mask[detectable_bins] = 1
        # Find bins which are detectable in the diagonal (intersect of
        # hori and verti)
        detect_mask_h = detect_mask[: (mat_n - diag)]
        detect_mask_v = detect_mask[mat_n - (mat_n - diag) :]
        detect_mask_diag = detect_mask_h & detect_mask_v
        detect_diag = matrix.diagonal(diag)[detect_mask_diag]
        dist[diag] = fun(detect_diag[detect_diag > 0])
    # Smooth the curve using isotonic regression: Find closest approximation
    # with the condition that point n+1 cannot be higher than point n.
    # (i.e. contacts can only decrease when increasing distance)
    if smooth and mat_n > 2:
        ir = IsotonicRegression(increasing=False)
        dist[~np.isfinite(dist)] = 0
        dist = ir.fit_transform(range(len(dist)), dist)

    return dist


def get_detectable_bins(mat, n_mads=3, inter=False):
    """
    Returns lists of detectable indices after excluding low interacting bins
    based on the proportion of zero pixel values in the matrix bins.

    Parameters
    ----------
    mat : scipy.sparse.coo_matrix
        A Hi-C matrix in tihe form of a 2D numpy array or coo matrix
    n_mads : int
        Number of median absolute deviation below the median required to
        consider bins non-detectable.
    inter : bool
        Whether the matrix is interchromosomal. Default is to consider the
        matrix is intrachromosomal (i.e. upper symmetric).

    Returns
    -------
    numpy ndarray :
        tuple of 2 1D arrays containing indices of low interacting rows and
        columns, respectively.
    """
    matrix = mat.copy()
    matrix.eliminate_zeros()

    def mad(x): return ss.median_absolute_deviation(x, nan_policy="omit")

    if not inter:
        if matrix.shape[0] != matrix.shape[1]:
            raise ValueError("Intrachromosomal matrices must be symmetric.")
        # Replace nonzero pixels by ones to work on prop. of nonzero pixels
        matrix.data = np.ones(matrix.data.shape)
        # Compute number of nonzero values in each bin
        sum_bins = sum_mat_bins(matrix)
        # Compute variation in the number of nonzero pixels
        sum_mad = mad(sum_bins)
        # Find poor interacting rows and columns
        sum_med = np.median(sum_bins)
        detect_threshold = max(1, sum_med - sum_mad * n_mads)

        # Removal of poor interacting rows and columns
        good_bins = np.flatnonzero(sum_bins >= detect_threshold)
        good_bins = (good_bins, good_bins)
    else:
        # Adapted for asymetric matrices (need to compute rows and columns)
        sum_rows, sum_cols = matrix.sum(axis=1).A1, matrix.sum(axis=0).A1
        mad_rows, mad_cols = mad(sum_rows), mad(sum_cols)
        med_rows, med_cols = np.median(sum_rows), np.median(sum_cols)
        detect_threshold_rows = max(1, med_rows - mad_rows * n_mads)
        detect_threshold_cols = max(1, med_cols - mad_cols * n_mads)
        good_rows = np.flatnonzero(sum_rows > detect_threshold_rows)
        good_cols = np.flatnonzero(sum_cols > detect_threshold_cols)
        good_bins = (good_rows, good_cols)
    return good_bins


def detrend(
    matrix,
    detectable_bins=None,
    max_dist=None,
    smooth=False,
    fun=np.nanmean,
    max_val=10,
):
    """
    Detrends a Hi-C matrix by the distance law.
    The input matrix should have been normalised beforehandand.

    Parameters
    ----------
    matrix : scipy.sparse.csr_matrix
        The normalised intrachromosomal Hi-C matrix to detrend.
    detectable_bins : tuple
        Tuple containing a list of detectable rows and a list of columns on
        which to perform detrending. Poorly interacting indices have been
        excluded.
    max_dist : int
        Maximum number of bins from the diagonal at which to compute trend.
    smooth : bool
        Whether to use isotonic regression to smooth the trend.
    fun : callable
        Function to use on each diagonal to compute the trend.
    max_val : float or None
        Maximum value in the detrended matrix. Set to None to disable

    Returns
    -------
    numpy.ndarray :
        The detrended Hi-C matrix.
    """
    matrix = matrix.tocsr()
    y = distance_law(
        matrix,
        detectable_bins=detectable_bins,
        max_dist=max_dist,
        smooth=smooth,
        fun=fun,
    )
    y[np.isnan(y)] = 0.0
    # Detrending by the distance law
    clean_mat = matrix.tocoo()
    # clean_mat.data /= y_savgol[abs(clean_mat.row - clean_mat.col)]
    try:
        clean_mat.data = clean_mat.data / y[abs(clean_mat.row - clean_mat.col)]
    # If no nonzero value in matrix, do nothing
    except TypeError:
        pass
    clean_mat = clean_mat.tocsr()
    if max_val is not None:
        clean_mat[clean_mat >= max_val] = 1
    return clean_mat


def ztransform(matrix):
    """
    Z transformation for Hi-C matrices.

    Parameters
    ----------
    matrix : scipy.sparse.coo_matrix
        A Hi-C matrix in sparse format.

    Returns
    -------
    scipy.sparse.coo_matrix:
        The detrended Hi-C matrix
    """

    mat = matrix.copy()
    mu = np.mean(mat.data)
    sd = np.std(mat.data)
    mat.data -= mu
    mat.data /= sd

    return mat


def sum_mat_bins(mat):
    """
    Compute the sum of matrices bins (i.e. rows or columns) using
    only the upper triangle, assuming symmetrical matrices.

    Parameters
    ----------
    mat : scipy.sparse.coo_matrix
        Contact map in sparse format, either in upper triangle or
        full matrix.

    Returns
    -------
    numpy.ndarray :
        1D array of bin sums.
    """
    # Equivalaent to row or col sum on a full matrix
    # Note: mat.sum returns a 'matrix' object. A1 extracts the 1D flat array
    # from the matrix
    return mat.sum(axis=0).A1 + mat.sum(axis=1).A1 - mat.diagonal(0)


def subsample_contacts(M, n_contacts):
    """Bootstrap sampling of contacts in a sparse Hi-C map.

    Parameters
    ----------
    M : scipy.sparse.coo_matrix
        The input Hi-C contact map in sparse format.
    n_contacts : int
        The number of contacts to sample.

    Returns
    -------
    scipy.sparse.coo_matrix
        A new matrix with a fraction of the original contacts.
    """

    S = M.data.copy()
    # Match cell idx to cumulative number of contacts
    cum_counts = np.cumsum(S)
    # Total number of contacts to sample
    tot_contacts = int(cum_counts[-1])

    # Sample desired number of contacts from the range(0, n_contacts) array
    sampled_contacts = np.random.choice(
        int(tot_contacts), size=(n_contacts), replace=False
    )

    # Get indices of sampled contacts in the cum_counts array
    idx = np.searchsorted(cum_counts, sampled_contacts, side="right")

    # Bin those indices to the same dimensions as matrix data to get counts
    sampled_counts = np.bincount(idx, minlength=S.shape[0])

    # Get nonzero values to build new sparse matrix
    nnz_mask = sampled_counts > 0
    sampled_counts = sampled_counts[nnz_mask].astype(np.float64)
    sampled_rows = M.row[nnz_mask]
    sampled_cols = M.col[nnz_mask]

    return sp.coo_matrix(
        (sampled_counts, (sampled_rows, sampled_cols)),
        shape=(M.shape[0], M.shape[1]),
    )


def frame_missing_mask(mask, kernel_shape, sym_upper=False, max_dist=None):
    """
    Adds a frame around input mask, given a kernel. The goal of this
    frame is define margins around the matrix where the kernel will not perform
    convolution (denoted by 1). If the matrix is upper symmetric, a margin of
    half the kernel's width is added below the diagonal and a maximum distance
    from the diagonal above which margins need not be drawn can be considered.
    Otherwise Margins are simply added on all 4 sides of the matrix.
    
    ::

        signal    kernel   _________
        ______   ____      |#######|
        |     |  |   | ==> |#     #|
        |     |  |___|     |#     #|
        |     |            |#     #|
        |_____|            |#     #|
                           |#######|
                           --------

    Parameters
    ----------
    mask : scipy.sparse.csr_matrix of bool
        The mask around which to add margins.
    kernels_shape : tuple of ints
        The number of rows and kernel in the input kernel. Margins will be half
        these values.
    sym_upper : bool
        Whether the signal is a symmetric upper triangle matrix. If so, values
        on a margin below the diagonal will be masked.
    max_dist : int or None
        Number of diagonals to keep

    Returns
    -------
    framed_mask : scipy.sparse.csr_matrix of bool
        The input mask with a padding of True around the edges. If sym_upper
        is True, a padding is also added below the diagonal.
    """
    if mask.dtype != bool:
        raise ValueError("Mask must contain boolean values")
    if not sp.issparse(mask):
        raise ValueError("Mask must be a sparse matrix")

    framed_mask = mask.copy()
    ms, ns = mask.shape
    mk, nk = kernel_shape
    if sym_upper and (max_dist is not None):
        # Remove diagonals further than scan distance in the input mask
        framed_mask = diag_trim(framed_mask, max_dist + max(nk, mk)).tocsr()
        max_m = max_dist + mk
        max_n = max_dist + nk
    else:
        max_m, max_n = ms, ns
    # Up and down margins initialized with zeros and filled as needed
    margin_1 = sp.csr_matrix((mk - 1, ns), dtype=bool)
    margin_2 = sp.csr_matrix((mk - 1, ns), dtype=bool)
    if sym_upper and (max_dist is not None):
        # Margin 1 (top) is in upper triangle -> fill missing up to scan dist
        margin_1[:, :max_n] = 1
    else:
        margin_1[:, :] = 1
        margin_2[:, :] = 1
    framed_mask = sp.vstack([margin_1, framed_mask, margin_2], format="csr")

    # Left and right
    margin_1 = sp.csr_matrix((ms + 2 * (mk - 1), nk - 1), dtype=bool)
    margin_2 = sp.csr_matrix((ms + 2 * (mk - 1), nk - 1), dtype=bool)

    if sym_upper and (max_dist is not None):
        # Margin 2 (right) is in upper triangle-> fill missing up to scan dist
        margin_2[-(max_m + 1) :, :] = 1
        # Fill only the start of left margin for the top-left corner
        margin_1[: mk - 1, :] = 1
    else:
        margin_1[:, :] = 1
        margin_2[:, :] = 1
    framed_mask = sp.hstack([margin_1, framed_mask, margin_2], format="csr")

    if sym_upper:
        # LIL format is much faster when changing sparsity
        framed_mask = framed_mask.tolil()
        # Add margin below diagonal
        big_k = max(nk, mk)
        dia_margins = np.ones(big_k)
        dia_offsets = np.arange(-1, -big_k-1, -1)
        framed_mask += sp.diags(
            dia_margins,
            dia_offsets,
            shape=framed_mask.shape,
            format="lil",
            dtype=bool,
        )
        framed_mask = framed_mask.tocsr()
    return framed_mask


def check_missing_mask(signal, mask):
    """
    Ensure all elements defined as missing by the mask are set to zero in the
    signal. If this is not the case, raises an error.

    Parameters
    ----------
    signal : numpy.ndarray of floats or scipy.sparse.csr_matrix of floats
        The signal to be checked.
    mask : numpy.ndarray of bools or scipy.sparse.csr_matrix of bools
        The mask defining missing values as True and valid values as False.
    """

    if sp.issparse(mask):
        # Check if there are nonzero values in the signal reported as missing
        # by the mask
        missing_with_signal = np.nonzero(
            abs(signal[mask.nonzero()[0], mask.nonzero()[1]]) > 0
        )[0]
        if len(missing_with_signal) > 0:
            raise ValueError(
                "There are",
                len(missing_with_signal),
                "non-zero elements reported as missing.",
            )
    else:
        if np.sum(abs(signal[mask > 0])) > 1e-10:
            raise ValueError(
                "There are",
                str(np.sum(abs(signal[mask > 0]))),
                "non-zero elements reported as missing.",
            )


def make_missing_mask(
    shape, valid_rows, valid_cols, max_dist=None, sym_upper=False
):
    """
    Given lists of valid rows and columns, generate a sparse matrix mask with
    missing pixels denoted as 1 and valid pixels as 0. If a max_dist is
    provided, upper symmetric matrices will only be flagged up to max_dist
    pixels from the diagonal.

    Parameters
    ----------
    shape : tuple of ints
        Shape of the mask to generate.
    valid_rows : numpy.ndarray of ints
        Array with the indices of valid rows that should be set to 0 in the
        mask.
    valid_cols : numpy.ndarray of ints
        Array with the indices of valid rows that should be set to 0 in the
        mask.
    max_dist : int or None
        The maximum diagonal distance at which masking should take place.
    sym_upper : bool
        Whether the matrix is symmetric upper. If so, max_dist is ignored

    Returns
    -------
    scipy.sparse.csr_matrix of bool
        The mask containing False values where pixels are valid and True valid
        where pixels are missing
    """
    # Error if the matrix upper symmetric but shape is rectangle or missing
    # rows and cols are different
    sm, sn = shape
    if sym_upper and (sm != sn or len(valid_rows) != len(valid_cols)):
        raise ValueError("Rectangular matrices cannot be upper symmetric")

    # Get a boolean array of missing (1) and valid (0) rows
    missing_rows = valid_to_missing(valid_rows, sm)
    # When matrix is sym., rows and cols are synonym, no need to compute 2x
    if sym_upper:
        missing_cols = missing_rows
    else:
        missing_cols = valid_to_missing(valid_cols, sn)
    # If upper sym., fill only upper diag up to max_dist.
    # E. g. with bins 1 and 3 missing
    # and a max_dist of 1:
    # 0 1 0 0 0
    # 0 1 1 0 0
    # 0 0 0 1 0
    # 0 0 0 1 1
    # 0 0 0 0 0
    # For each missing bin, mask is apply 1 pixel upwards and 1 to the right
    # to fill only the upper triangle up to max_dist
    if sym_upper:
        # If no max dist has been specified, fill the whole upper triangle
        if max_dist is None:
            max_dist = min(shape)
        # Generate matrix of distance shifts by row.
        # Shape is len(missing_rows) x (max_dist + 1)
        # e.g.: 2 missing rows and max dist of 1
        # 0 0
        # 1 1
        row_shifts = np.tile(
            np.array(range(max_dist + 1)), (len(missing_rows), 1)
        ).T
        # Compute row positions upwards to diagonal by subtracting missing rows
        # to the shifts. Following the previous example, if missing rows are
        # bins 1 and 3:
        #  1 3
        #  0 2
        rows_before = (missing_rows - row_shifts).flatten("F")
        # looking at pixels up from the bins, cols remain the same:
        # 1 3
        # 1 3
        cols_before = np.repeat(missing_rows, max_dist+1)
        # Compute col position to the right until diagonal by adding the shift
        # Note: upper symmetric, so row_shifts = col_shift_
        # 1 3
        # 2 4
        cols_after = (missing_cols + row_shifts).flatten("F")
        # This time, rows remain constant since we are computing positions to
        # the right
        rows_after = np.repeat(missing_cols, max_dist+1)
        # Combine positions to the right and upwards
        rows = np.concatenate([rows_before, rows_after])
        cols = np.concatenate([cols_before, cols_after])
        data = np.ones(rows.shape, dtype=bool)
        # Remove entries where rows or cols are negative or larger than shape
        valid = (cols < sm) & (cols >= 0) & (rows < sm) & (rows >= 0)
        # Build mask mat with miss bins up to max scan dist in upper triangle
        mask = sp.coo_matrix(
            (data[valid], (rows[valid], cols[valid])), shape=shape, dtype=bool
        ).tocsr()
    else:
        mask = sp.csr_matrix(shape, dtype=bool)
        mask[missing_rows, :] = 1
        mask[:, missing_cols] = 1

    return mask


def zero_pad_sparse(mat, margin_h, margin_v, fmt="coo"):
    """
    Adds margin of zeros around an input sparse matrix.

    Parameters
    ----------
    mat : scipy.sparse.csr_matrix
        The matrix to be padded.
    margin_h : int
        The width of the horizontal margin to add on the left and right of the
        matrix.
    margin_v : int
        The width of the vertical margin to add on the top and bottom of the
        matrix.
    fmt : string
        The desired scipy sparse format of the output matrix

    Returns
    -------
    scipy.sparse.csr_matrix :
        The padded matrix of dimensions (m + 2 * margin_h, n + 2 * margin_v).

    Examples
    --------
    >>> m = sp.csr_matrix(np.array([[1, 2], [10, 20]]))
    >>> zero_pad_sparse(m, 2, 1).toarray()
    array([[ 0,  0,  0,  0,  0,  0],
           [ 0,  0,  1,  2,  0,  0],
           [ 0,  0, 10, 20,  0,  0],
           [ 0,  0,  0,  0,  0,  0]])
    """

    sm, sn = mat.shape
    padded_mat = mat.copy()
    # Up and down margins initialized with zeros and filled as needed
    margin_h_0 = sp.csr_matrix((sm, margin_h), dtype=mat.dtype)
    margin_v_0 = sp.csr_matrix((margin_v, sn + 2 * margin_h), dtype=mat.dtype)
    padded_mat = sp.hstack([margin_h_0, padded_mat, margin_h_0], format="csr")
    padded_mat = sp.vstack([margin_v_0, padded_mat, margin_v_0], format="csr")

    return padded_mat


def crop_kernel(kernel, target_size):
    """
    Crop a kernel matrix to target size horizontally and vertically.
    If the target size is even, the target size is adjusted to the
    next integer up.

    Parameters
    ----------
    kernel : numpy.ndarray of floats
        Image to crop.
    target_size : tuple of ints
        Tuple defining the target shape of the kernel, takes the
        form (rows, cols) where rows and cols are odd numbers.

    Returns
    -------
    cropped : numpy.ndarray of floats
        New image no larger than target dimensions
    """
    # Use list for mutability
    target = [d for d in target_size]
    adjusted = False
    for dim in range(len(target)):
        if not target[dim] % 2:
            target[dim] += 1
            adjusted = True
    if adjusted:
        sys.stderr.write(
            "WARNING: Cropped kernel size adjusted to "
            f"{target[0]}x{target[1]} to keep odd dimensions.\n"
        )

    source_m, source_n = kernel.shape
    target_m, target_n = target
    # Define horizontal and vertical margins to trim
    if source_m > target_m:
        margin_rows = (source_m - target_m) // 2
    else:
        margin_rows = 0
    if source_n > target_n:
        margin_cols = (source_n - target_n) // 2
    else:
        margin_cols = 0

    cropped = kernel[
        margin_rows : (source_m - margin_rows),
        margin_cols : (source_n - margin_cols),
    ]

    return cropped


def resize_kernel(
    kernel,
    kernel_res=None,
    signal_res=None,
    factor=None,
    min_size=7,
    quiet=False,
):
    """
    Resize a kernel matrix based on the resolution at which it was defined and
    the signal resolution. E.g. if a kernel matrix was generated for 10kb and
    the input signal is 20kb, kernel size will be divided by two. If the kernel
    is enlarged, pixels are interpolated with a spline of degree 1.
    Alternatively, a resize factor can be provided. In the example above, the
    factor would be 0.5.

    Parameters
    ----------
    kernel : numpy.ndarray
        Kernel matrix.
    kernel_res : int
        Resolution for which the kernel was designed. Mutually exclusive with
        factor.
    signal_res : int
        Resolution of the signal matrix in basepair per matrix bin. Mutually
        exclusive with factor.
    factor : float
        Resize factor. Can be provided as an alternative to kernel_res and
        signal_res. Values above 1 will enlarge the kernel, values below 1 will
        shrink it.
    min_size : int
        Lower bound, in number of rows/column allowed when resizing the kernel.
    quiet : bool
        Suppress warnings if resize factor was adjusted.

    Returns
    -------
    resized_kernel : numpy.ndarray
        The resized input kernel.
    """
    km, kn = kernel.shape
    if km != kn:
        raise ValueError("kernel must be square.")
    if not (km % 2) or not (kn % 2):
        raise ValueError("kernel size must be odd.")

    if factor is not None:
        if kernel_res is not None or signal_res is not None:
            raise ValueError(
                "factor is mutually exclusive with resolution "
                "parameters (kernel_res and signal_res)."
            )
        resize_factor = factor
    else:
        if kernel_res is None or signal_res is None:
            raise ValueError(
                "You must provide either a resize factor or the signal and "
                "kernel resolutions."
            )
        # Define by how many times kernel must be enlarged for its pixels to
        # match the signal's pixels
        resize_factor = kernel_res / signal_res
    if km * resize_factor < min_size:
        resize_factor = min_size / km

    resized_kernel = ndi.zoom(kernel, resize_factor, order=1)

    if not resized_kernel.shape[0] % 2:
        # Compute the factor required to yield a dimension smaller by one
        adj_resize_factor = (resized_kernel.shape[0] - 1) / km
        if not quiet:
            sys.stderr.write(
                f"Adjusting resize factor from {resize_factor} to {adj_resize_factor}.\n"
            )
        resized_kernel = ndi.zoom(kernel, adj_resize_factor, order=1)

    return resized_kernel


def factorise_kernel(kernel, prop_info=0.999):
    """
    Performs truncated SVD on an input kernel, returning the singular vectors
    necessary to retain a given proportion of information contained in the
    kernel.

    Parameters
    ----------
    kernel : numpy.ndarray of floats
        The input 2D kernel to factorise.
    prop_info : float
        Proportion of information to retain.

    Returns
    -------
    tuple of numpy.ndarrays of floats
        A tuple containing the truncated left and right singular matrices,
        where each singular vector has been multiplied by the square root of
        their respective singular values.
    """
    u, sigma, v = la.svd(kernel)
    total_info = np.sum(sigma ** 2)
    # Compute min. number of singular vectors to retain enough info
    keep_k = np.flatnonzero(np.cumsum(sigma ** 2) > prop_info * total_info)[0] + 1
    if keep_k > np.floor(min(kernel.shape) / 2):
        sys.stderr.write(
            f"Warning: Kernel factorisation required {keep_k} singular,"
            "vectors this may result in slow operations.\n",
        )
    # Truncate singular matrix to the keep only required vectors
    u = u[:, :keep_k]
    v = v[:keep_k, :]
    # Multiply each singular vector by the sqrt of its singular value
    for i in range(keep_k):
        u[:, i] *= np.sqrt(sigma[i])
        v[i, :] *= np.sqrt(sigma[i])

    return (u, v)


def valid_to_missing(valid, size):
    """
    Given an array of valid indices, return the corrsesponding array of missing
    indices.

    Parameters
    ---------
    valid : numpy.ndarray of ints
        The valid indices.
    size : int
        The size of the matrix (maximum possible index + 1).

    Return
    ------
    missing : numpy.ndarray of ints
        The missing indices.

    """
    missing = np.ones(size, dtype=bool)
    try:
        missing[valid] = False
    # In case there is no valid index
    except IndexError:
        pass
    missing = np.flatnonzero(missing)
    return missing
